from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import pytest

import DIRAC
import six


@pytest.mark.parametrize(
    "path, expected",
    [
        ("/opt/dirac/versions/v7.2.1-1622561290/Linux-x86_64/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.2.1-1622561290/Linux-x86_64/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.2.0a14.dev11+ga5b9ec512-1622561290/Linux-x86_64/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.200.999-1622561290/Linux-x86_64/", "/opt/dirac"),
        ("/opt/dirac/versions/v487.200.999-1622561290/Linux-x86_64/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.2.1a1-1622561290/Linux-x86_64/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.2.1a12-1622561290/Linux-x86_64/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.2.1b10-1622561290/Linux-x86_64/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.2.1rc2-1622561290/Linux-x86_64/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.2.1-1622561290/Linux-ppc64le/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.2.1-1622561290/Linux-aarch64/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.2.1-1622561290/Darwin-x86_64/", "/opt/dirac"),
        ("/opt/dirac/versions/v10.2.1-1622561290/Darwin-arm64/", "/opt/dirac"),
        ("/cvmfs/lhcb.cern.ch/lhcbdirac/versions/v10.2.1-1622561290/Linux-x86_64/", "/cvmfs/lhcb.cern.ch/lhcbdirac"),
        ("/cvmfs/lhcb.cern.ch/lhcbdirac/versions/v10.2.1a1-1622561290/Linux-x86_64/", "/cvmfs/lhcb.cern.ch/lhcbdirac"),
        ("/diracos/", "/diracos"),
        ("/", "/"),
        ("/opt/dirac/", "/opt/dirac"),
        ("/opt/versions/", "/opt/versions"),
        ("/opt/versions/1234", "/opt/versions/1234"),
        ("/opt/versions/diracos", "/opt/versions/diracos"),
    ],
)
@pytest.mark.skipif(six.PY2, reason="Only makes sense for Python 3 installs")
def test_computeRootPath(path, expected):
    assert DIRAC._computeRootPath(path) == expected
    # Trailing / shouldn't affect the rootPath
    assert DIRAC._computeRootPath(path + "//") == expected
    if path != "/":
        assert DIRAC._computeRootPath(path.rstrip("/")) == expected



# pylint: disable=protected-access
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import math
import operator
from functools import reduce

from PIL import Image
import six


def compare(file1Path, file2Path):
    """
    Function used to compare two plots
    returns 0.0 if both are identical

    :param str file1Path: Path to the file1.
    :param str file2Path: Path to the file2.

    :return: float value rms.
    """
    if six.PY3:
        file2Path += ".py3k"

    # Crops image to remove the "Generated on xxxx UTC" string
    image1 = Image.open(file1Path).crop((0, 0, 800, 570))
    image2 = Image.open(file2Path).crop((0, 0, 800, 570))

    h1 = image1.histogram()
    h2 = image2.histogram()
    rms = math.sqrt(reduce(operator.add, map(lambda a, b: (a - b) ** 2, h1, h2)) / len(h1))
    return rms



""" Accounting Type for Pilot Submission

    Filled by the "WorkloadManagement/SiteDirector" agent(s)
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from DIRAC.AccountingSystem.Client.Types.BaseAccountingType import BaseAccountingType

__RCSID__ = "$Id$"


class PilotSubmission(BaseAccountingType):
    """Accounting Type class for Pilot Submission"""

    def __init__(self):
        super(PilotSubmission, self).__init__()

        self.definitionKeyFields = [
            ("HostName", "VARCHAR(100)"),
            ("SiteDirector", "VARCHAR(100)"),
            ("Site", "VARCHAR(100)"),
            ("CE", "VARCHAR(100)"),
            ("Queue", "VARCHAR(100)"),
            ("Status", "VARCHAR(100)"),
        ]
        self.definitionAccountingFields = [("NumTotal", "INT UNSIGNED"), ("NumSucceeded", "INT UNSIGNED")]

        self.bucketsLength = [
            (86400 * 2, 900),  # <2d = 15m
            (86400 * 10, 9000),  # <10d = 2.5h
            (86400 * 35, 18000),  # <35d = 5h
            (86400 * 30 * 6, 86400),  # >5d <6m = 1d
            (86400 * 600, 604800),
        ]  # >6m = 1w

        self.dataTimespan = 86400 * 30 * 14  # Only keep the last 14 months of data
        self.checkType()



""" Test class for Stalled Job Agent
"""

# imports
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import pytest
from mock import MagicMock

# DIRAC Components
from DIRAC.WorkloadManagementSystem.Agent.StalledJobAgent import StalledJobAgent
from DIRAC import gLogger

# Mock Objects
mockAM = MagicMock()
mockNone = MagicMock()
mockNone.return_value = None

gLogger.setLevel("DEBUG")


@pytest.fixture
def sja(mocker):
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.StalledJobAgent.AgentModule.__init__")
    mocker.patch(
        "DIRAC.WorkloadManagementSystem.Agent.StalledJobAgent.AgentModule._AgentModule__moduleProperties",
        side_effect=lambda x, y=None: y,
        create=True,
    )
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.StalledJobAgent.JobDB.__init__", side_effect=mockNone)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.StalledJobAgent.JobLoggingDB.__init__", side_effect=mockNone)
    mocker.patch("DIRAC.WorkloadManagementSystem.Agent.StalledJobAgent.getSystemInstance", side_effect=mockNone)

    stalledJobAgent = StalledJobAgent()
    stalledJobAgent._AgentModule__configDefaults = mockAM
    stalledJobAgent.log = gLogger
    stalledJobAgent.initialize()
    stalledJobAgent.jobDB.log = gLogger
    stalledJobAgent.log.setLevel("DEBUG")
    stalledJobAgent.stalledTime = 120

    return stalledJobAgent


def test__sjaFunctions(sja):
    """Testing StalledJobAgent()"""

    assert not sja._failSubmittingJobs()["OK"]
    assert not sja._kickStuckJobs()["OK"]
    assert not sja._failStalledJobs(0)["OK"]
    assert not sja._markStalledJobs(0)["OK"]



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-job-logging-info
# Author :  Stuart Paterson
########################################################################
"""
Retrieve history of transitions for a DIRAC job

Example:
  $ dirac-wms-job-logging-info 1
  Status                        MinorStatus                         ApplicationStatus             DateTime
  Received                      Job accepted                        Unknown                       2011-02-14 10:12:40
  Received                      False                               Unknown                       2011-02-14 11:03:12
  Checking                      JobSanity                           Unknown                       2011-02-14 11:03:12
  Checking                      JobScheduling                       Unknown                       2011-02-14 11:03:12
  Waiting                       Pilot Agent Submission              Unknown                       2011-02-14 11:03:12
  Matched                       Assigned                            Unknown                       2011-02-14 11:27:17
  Matched                       Job Received by Agent               Unknown                       2011-02-14 11:27:27
  Matched                       Submitted To CE                     Unknown                       2011-02-14 11:27:38
  Running                       Job Initialization                  Unknown                       2011-02-14 11:27:42
  Running                       Application                         Unknown                       2011-02-14 11:27:48
  Completed                     Application Finished Successfully   Unknown                       2011-02-14 11:28:01
  Completed                     Uploading Output Sandbox            Unknown                       2011-02-14 11:28:04
  Completed                     Output Sandbox Uploaded             Unknown                       2011-02-14 11:28:07
  Done                          Execution Complete                  Unknown                       2011-02-14 11:28:07
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["JobID:    DIRAC Job ID"])
    _, args = Script.parseCommandLine(ignoreErrors=True)

    from DIRAC.Interfaces.API.Dirac import Dirac, parseArguments

    dirac = Dirac()
    exitCode = 0
    errorList = []

    for job in parseArguments(args):

        result = dirac.getJobLoggingInfo(job, printOutput=True)
        if not result["OK"]:
            errorList.append((job, result["Message"]))
            exitCode = 2

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" :mod: RequestTests
    =======================

    .. module: RequestTests
    :synopsis: test cases for Request class
    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    test cases for Request class
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import datetime

import six

from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.File import File
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.ReqClient import printRequest


def optimizeRequest(req, printOutput=None):
    from DIRAC import gLogger

    if printOutput:
        if isinstance(printOutput, six.string_types):
            gLogger.always("Request %s:" % printOutput)
        printRequest(req)
        gLogger.always("=========== Optimized ===============")
    res = req.optimize()
    if printOutput:
        printRequest(req)
        gLogger.always("")
    return res


def createRequest(reqType):
    r = Request()

    # Simple failover
    op1 = Operation()
    f = File()
    f.LFN = "/This/is/an/LFN"
    op1.addFile(f)
    op1.Type = "ReplicateAndRegister"
    op1.SourceSE = "CERN-FAILOVER"
    op1.TargetSE = "CERN-BUFFER"
    r.addOperation(op1)

    # You cannot reuse the same File object,
    # since it is a different entry in the DB
    fr = File()
    fr.LFN = "/This/is/an/LFN"
    op2 = Operation()
    op2.addFile(fr)
    op2.Type = "RemoveReplica"
    op2.TargetSE = "CERN-FAILOVER"
    r.addOperation(op2)
    if reqType == 0:
        return r

    # two files for Failover
    f1 = File()
    f1.LFN = "/This/is/a/second/LFN"
    op3 = Operation()
    op3.addFile(f1)
    op3.Type = "ReplicateAndRegister"
    op3.SourceSE = "CERN-FAILOVER"
    op3.TargetSE = "CERN-BUFFER"
    r.addOperation(op3)

    f1r = File()
    f1r.LFN = "/This/is/a/second/LFN"
    op3 = Operation()
    op3.addFile(f1r)
    op3.Type = "RemoveReplica"
    op3.TargetSE = "CERN-FAILOVER"
    r.addOperation(op3)
    if reqType == 1:
        return r

    op = Operation()
    op.Type = "ForwardDiset"
    if reqType == 2:
        r.addOperation(op)
        return r

    r.insertBefore(op, r[0])
    if reqType == 3:
        return r

    op4 = Operation()
    op4.Type = "ForwardDiset"
    r.addOperation(op4)
    if reqType == 4:
        return r

    # 2 different FAILOVER SEs: removal not optimized
    r[1].SourceSE = "RAL-FAILOVER"
    r[2].SourceSE = "RAL-FAILOVER"
    if reqType == 5:
        return r

    # 2 different destinations, same FAILOVER: replication not optimized
    r[3].SourceSE = "RAL-FAILOVER"
    r[4].SourceSE = "RAL-FAILOVER"
    r[3].TargetSE = "RAL-BUFFER"
    if reqType == 6:
        return r

    print("This should not happen, reqType =", reqType)


########################################################################


def test_CtorSerilization():
    """c'tor and serialization"""
    req = Request()
    assert isinstance(req, Request)
    assert req.JobID == 0
    assert req.Status == "Waiting"

    req = Request({"RequestName": "test", "JobID": 12345})
    assert isinstance(req, Request)
    assert req.RequestName == "test"
    assert req.JobID == 12345
    assert req.Status == "Waiting"

    req.SourceComponent = "test component"
    assert req.SourceComponent == b"test component"

    toJSON = req.toJSON()
    assert toJSON["OK"], "JSON serialization failed"

    fromJSON = toJSON["Value"]
    req = Request(fromJSON)


def test_Props():
    """props"""
    # # valid values
    req = Request()

    req.RequestID = 1
    assert req.RequestID == 1

    req.RequestName = "test"
    assert req.RequestName == "test"

    req.JobID = 1
    assert req.JobID == 1

    req.CreationTime = "1970-01-01 00:00:00"
    assert req.CreationTime == datetime.datetime(1970, 1, 1, 0, 0, 0)
    req.CreationTime = datetime.datetime(1970, 1, 1, 0, 0, 0)
    assert req.CreationTime == datetime.datetime(1970, 1, 1, 0, 0, 0)

    req.SubmitTime = "1970-01-01 00:00:00"
    assert req.SubmitTime == datetime.datetime(1970, 1, 1, 0, 0, 0)
    req.SubmitTime = datetime.datetime(1970, 1, 1, 0, 0, 0)
    assert req.SubmitTime == datetime.datetime(1970, 1, 1, 0, 0, 0)

    req.LastUpdate = "1970-01-01 00:00:00"
    assert req.LastUpdate == datetime.datetime(1970, 1, 1, 0, 0, 0)
    req.LastUpdate = datetime.datetime(1970, 1, 1, 0, 0, 0)
    assert req.LastUpdate == datetime.datetime(1970, 1, 1, 0, 0, 0)

    req.Error = ""


def test_Operations():
    """operations arithmetic and state machine"""
    req = Request()
    assert len(req) == 0

    transfer = Operation()
    transfer.Type = "ReplicateAndRegister"
    transfer.addFile(File({"LFN": "/a/b/c", "Status": "Waiting"}))

    getWaiting = req.getWaiting()
    assert getWaiting["OK"]
    assert getWaiting["Value"] is None

    req.addOperation(transfer)
    assert len(req) == 1
    assert transfer.Order == req.Order
    assert transfer.Status == "Waiting"

    getWaiting = req.getWaiting()
    assert getWaiting["OK"]
    assert getWaiting["Value"] == transfer

    removal = Operation({"Type": "RemoveFile"})
    removal.addFile(File({"LFN": "/a/b/c", "Status": "Waiting"}))

    req.insertBefore(removal, transfer)

    getWaiting = req.getWaiting()
    assert getWaiting["OK"]
    assert getWaiting["Value"] == removal

    assert len(req) == 2
    assert [op.Status for op in req] == ["Waiting", "Queued"]
    assert req.subStatusList() == ["Waiting", "Queued"]

    assert removal.Order == 0
    assert removal.Order == req.Order

    assert transfer.Order == 1

    assert removal.Status == "Waiting"
    assert transfer.Status == "Queued"

    for subFile in removal:
        subFile.Status = "Done"
    removal.Status = "Done"

    assert removal.Status == "Done"

    assert transfer.Status == "Waiting"
    assert transfer.Order == req.Order

    # len, looping
    assert len(req) == 2
    assert [op.Status for op in req] == ["Done", "Waiting"]
    assert req.subStatusList() == ["Done", "Waiting"]

    digest = req.toJSON()
    assert digest["OK"]

    getWaiting = req.getWaiting()
    assert getWaiting["OK"]
    assert getWaiting["Value"] == transfer


def test_FTS():
    """FTS state machine"""

    req = Request()
    req.RequestName = "FTSTest"

    ftsTransfer = Operation()
    ftsTransfer.Type = "ReplicateAndRegister"
    ftsTransfer.TargetSE = "CERN-USER"

    ftsFile = File()
    ftsFile.LFN = "/a/b/c"
    ftsFile.Checksum = "123456"
    ftsFile.ChecksumType = "Adler32"

    ftsTransfer.addFile(ftsFile)
    req.addOperation(ftsTransfer)

    assert req.Status == "Waiting", "1. wrong request status: %s" % req.Status
    assert ftsTransfer.Status == "Waiting", "1. wrong ftsStatus status: %s" % ftsTransfer.Status

    # # scheduled
    ftsFile.Status = "Scheduled"

    assert ftsTransfer.Status == "Scheduled", "2. wrong status for ftsTransfer: %s" % ftsTransfer.Status
    assert req.Status == "Scheduled", "2. wrong status for request: %s" % req.Status

    # # add new operation before FTS
    insertBefore = Operation()
    insertBefore.Type = "RegisterReplica"
    insertBefore.TargetSE = "CERN-USER"
    insertFile = File()
    insertFile.LFN = "/a/b/c"
    insertFile.PFN = "http://foo/bar"
    insertBefore.addFile(insertFile)
    req.insertBefore(insertBefore, ftsTransfer)

    assert insertBefore.Status == "Waiting", "3. wrong status for insertBefore: %s" % insertBefore.Status
    assert ftsTransfer.Status == "Scheduled", "3. wrong status for ftsStatus: %s" % ftsTransfer.Status
    assert req.Status == "Waiting", "3. wrong status for request: %s" % req.Status

    # # prev done
    insertFile.Status = "Done"

    assert insertBefore.Status == "Done", "4. wrong status for insertBefore: %s" % insertBefore.Status
    assert ftsTransfer.Status == "Scheduled", "4. wrong status for ftsStatus: %s" % ftsTransfer.Status
    assert req.Status == "Scheduled", "4. wrong status for request: %s" % req.Status

    # # reschedule
    ftsFile.Status = "Waiting"

    assert insertBefore.Status == "Done", "5. wrong status for insertBefore: %s" % insertBefore.Status
    assert ftsTransfer.Status == "Waiting", "5. wrong status for ftsStatus: %s" % ftsTransfer.Status
    assert req.Status == "Waiting", "5. wrong status for request: %s" % req.Status

    # # fts done
    ftsFile.Status = "Done"

    assert insertBefore.Status == "Done", "5. wrong status for insertBefore: %s" % insertBefore.Status
    assert ftsTransfer.Status == "Done", "5. wrong status for ftsStatus: %s" % ftsTransfer.Status
    assert req.Status == "Done", "5. wrong status for request: %s" % req.Status


def test_StateMachine():
    """state machine tests"""
    r = Request({"RequestName": "SMT"})
    assert r.Status == "Waiting", "1. wrong status %s" % r.Status

    r.addOperation(Operation({"Status": "Queued"}))
    assert r.Status == "Waiting", "2. wrong status %s" % r.Status

    r.addOperation(Operation({"Status": "Queued"}))
    assert r.Status == "Waiting", "3. wrong status %s" % r.Status

    r[0].Status = "Done"
    assert r.Status == "Waiting", "4. wrong status %s" % r.Status

    r[1].Status = "Done"
    assert r.Status == "Done", "5. wrong status %s" % r.Status

    r[0].Status = "Failed"
    assert r.Status == "Failed", "6. wrong status %s" % r.Status

    r[0].Status = "Queued"
    assert r.Status == "Waiting", "7. wrong status %s" % r.Status

    r.insertBefore(Operation({"Status": "Queued"}), r[0])
    assert r.Status == "Waiting", "8. wrong status %s" % r.Status

    r.insertBefore(Operation({"Status": "Queued"}), r[0])
    assert r.Status == "Waiting", "9. wrong status %s" % r.Status

    r.insertBefore(Operation({"Status": "Scheduled"}), r[0])
    assert r.Status == "Scheduled", "10. wrong status %s" % r.Status

    r.insertBefore(Operation({"Status": "Queued"}), r[0])
    assert r.Status == "Waiting", "11. wrong status %s" % r.Status

    r[0].Status = "Failed"
    assert r.Status == "Failed", "12. wrong status %s" % r.Status

    r[0].Status = "Done"
    assert r.Status == "Scheduled", "13. wrong status %s" % r.Status

    r[1].Status = "Failed"
    assert r.Status == "Failed", "14. wrong status %s" % r.Status

    r[1].Status = "Done"
    assert r.Status == "Waiting", "15. wrong status %s" % r.Status

    r[2].Status = "Scheduled"
    assert r.Status == "Scheduled", "16. wrong status %s" % r.Status

    r[2].Status = "Queued"
    assert r.Status == "Waiting", "17. wrong status %s" % r.Status

    r[2].Status = "Scheduled"
    assert r.Status == "Scheduled", "18. wrong status %s" % r.Status

    r = Request()
    for _ in range(5):
        r.addOperation(Operation({"Status": "Queued"}))

    r[0].Status = "Done"
    assert r.Status == "Waiting", "19. wrong status %s" % r.Status

    r[1].Status = "Done"
    assert r.Status == "Waiting", "20. wrong status %s" % r.Status

    r[2].Status = "Scheduled"
    assert r.Status == "Scheduled", "21. wrong status %s" % r.Status

    r[2].Status = "Done"
    assert r.Status == "Waiting", "22. wrong status %s" % r.Status


def test_List():
    """setitem, delitem, getitem and dirty"""

    r = Request()

    ops = [Operation() for i in range(5)]

    for op in ops:
        r.addOperation(op)

    for i, op in enumerate(ops):
        assert op == r[i], "__getitem__ failed"

    op = Operation()
    r[0] = op
    assert op == r[0], "__setitem__ failed"

    del r[0]
    assert len(r) == 4, "__delitem__ failed"


def test_Optimize():
    title = {
        0: "Simple Failover",
        1: "Double Failover",
        2: "Double Failover + ForwardDiset",
        3: "ForwardDiset + Double Failover",
        4: "ForwardDiset + Double Failover + ForwardDiset",
        5: "ForwardDiset + Double Failover (# Failover SE) + ForwardDiset",
        6: "ForwardDiset + Double Failover (# Destination SE) + ForwardDiset",
    }
    debug = False
    if debug:
        print("")
    for reqType in title:
        r = createRequest(reqType)
        res = optimizeRequest(r, printOutput=title[reqType] if (debug and debug == reqType) else False)
        assert res["OK"]
        assert res["Value"]
        if reqType in (0, 1):
            assert len(r) == 2, "Wrong number of operations: %d" % len(r)
            assert r[0].Type == "ReplicateAndRegister"
            assert r[1].Type == "RemoveReplica"
        if reqType == 1:
            assert len(r[0]) == 2, "Wrong number of files: %d" % len(r[0])
            assert len(r[1]) == 2, "Wrong number of files: %d" % len(r[1])
        elif reqType == 2:
            assert len(r) == 3, "Wrong number of operations: %d" % len(r)
            assert r[0].Type == "ReplicateAndRegister"
            assert r[1].Type == "RemoveReplica"
            assert r[2].Type == "ForwardDiset"
            assert len(r[0]) == 2, "Wrong number of files: %d" % len(r[0])
            assert len(r[1]) == 2, "Wrong number of files: %d" % len(r[1])
        elif reqType == 3:
            assert len(r) == 3, "Wrong number of operations: %d" % len(r)
            assert r[1].Type == "ReplicateAndRegister"
            assert r[2].Type == "RemoveReplica"
            assert r[0].Type == "ForwardDiset"
            assert len(r[1]) == 2, "Wrong number of files: %d" % len(r[0])
            assert len(r[2]) == 2, "Wrong number of files: %d" % len(r[1])
        elif reqType == 4:
            assert len(r) == 4, "Wrong number of operations: %d" % len(r)
            assert r[1].Type == "ReplicateAndRegister"
            assert r[2].Type == "RemoveReplica"
            assert r[0].Type == "ForwardDiset"
            assert r[3].Type == "ForwardDiset"
            assert len(r[1]) == 2, "Wrong number of files: %d" % len(r[0])
            assert len(r[2]) == 2, "Wrong number of files: %d" % len(r[1])
        elif reqType == 5:
            assert len(r) == 5, "Wrong number of operations: %d" % len(r)
            assert r[1].Type == "ReplicateAndRegister"
            assert r[2].Type == "RemoveReplica"
            assert r[3].Type == "RemoveReplica"
            assert r[0].Type == "ForwardDiset"
            assert r[4].Type == "ForwardDiset"
            assert len(r[1]) == 2, "Wrong number of files: %d" % len(r[0])
            assert len(r[2]) == 1, "Wrong number of files: %d" % len(r[1])
            assert len(r[3]) == 1, "Wrong number of files: %d" % len(r[1])
        elif reqType == 6:
            assert len(r) == 5, "Wrong number of operations: %d" % len(r)
            assert r[1].Type == "ReplicateAndRegister"
            assert r[2].Type == "ReplicateAndRegister"
            assert r[3].Type == "RemoveReplica"
            assert r[0].Type == "ForwardDiset"
            assert r[4].Type == "ForwardDiset"
            assert len(r[1]) == 1, "Wrong number of files: %d" % len(r[0])
            assert len(r[2]) == 1, "Wrong number of files: %d" % len(r[1])
            assert len(r[3]) == 2, "Wrong number of files: %d" % len(r[1])



# https://docs.aws.amazon.com/AmazonS3/latest/API/API_Operations.html
# https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.list_objects
"""
Configuration of an S3 storage
Like others, but in protocol S3 add:

* SecureConnection: true if https, false otherwise
* Aws_access_key_id
* Aws_secret_access_key

if the Aws variables are not defined, it will try to go throught the S3Gateway

The key of the objects are the LFN without trailing path.
The Path should be the BucketName

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"


import copy
import errno
import functools

import os
import requests


import boto3
from botocore.exceptions import ClientError

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Adler import fileAdler
from DIRAC.Core.Utilities.DErrno import cmpError
from DIRAC.Core.Utilities.Pfn import pfnparse
from DIRAC.DataManagementSystem.Client.S3GatewayClient import S3GatewayClient
from DIRAC.Resources.Storage.StorageBase import StorageBase


LOG = gLogger.getSubLogger(__name__)


def _extractKeyFromS3Path(meth):
    """Decorator to split an s3 "external" url (s3://server:port/bucket/path)
    and return only the path part.
    """

    @functools.wraps(meth)
    def extractKey(self, urls, *args, **kwargs):

        # If set to False, we are already working with keys, so
        # skip all the splitting
        extractKeys = kwargs.pop("extractKeys", True)

        keysToUrls = {}
        keyArgs = {}

        successful = {}
        failed = {}

        if extractKeys:
            for url in urls:
                res = self._getKeyFromURL(url)  # pylint: disable=protected-access
                if not res["OK"]:
                    failed[url] = res["Message"]
                    continue

                key = res["Value"]
                keysToUrls[key] = url
                keyArgs[key] = urls[url]
        else:
            keyArgs = copy.copy(urls)

        result = meth(self, keyArgs, *args, **kwargs)

        if not result["OK"]:
            return result
        # Restore original paths

        for key in result["Value"]["Failed"]:
            failed[keysToUrls.get(key, key)] = result["Value"]["Failed"][key]
        for key in result["Value"]["Successful"]:
            successful[keysToUrls.get(key, key)] = result["Value"]["Successful"][key]

        result["Value"].update({"Successful": successful, "Failed": failed})
        return result

    return extractKey


class S3Storage(StorageBase):
    """
    .. class:: StorageBase

    """

    pluginName = "S3"

    _OUTPUT_PROTOCOLS = ["file", "s3", "http", "https"]

    def __init__(self, storageName, parameters):

        super(S3Storage, self).__init__(storageName, parameters)

        self.isok = True

        aws_access_key_id = parameters.get("Aws_access_key_id")
        aws_secret_access_key = parameters.get("Aws_secret_access_key")
        self.secureConnection = parameters.get("SecureConnection", "True") == "True"
        proto = "https" if self.secureConnection else "http"
        port = int(parameters.get("Port"))
        if not port:
            port = 443 if self.secureConnection else 80
        endpoint_url = "%s://%s:%s" % (proto, parameters["Host"], port)
        self.bucketName = parameters["Path"]

        self.s3_client = boto3.client(
            "s3",
            endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )

        self.srmSpecificParse = False
        self.pluginName = "S3"

        # if we have the credentials loaded, we can perform direct access
        # otherwise we have to go through the S3Gateway
        self.directAccess = aws_access_key_id and aws_secret_access_key
        self.S3GatewayClient = S3GatewayClient()

    def _getKeyFromURL(self, url):
        """Extract the Key from the URL.
        The key is basically the LFN without trailing slash

        I despise such path mangling, expecially after all the efforts to
        get ride of such method. However, since in the case of S3 we need
        to go back and forth between URL and LFN (for finding keys, checking
        accesses in the gw, etc), there is no other option...

        :param url: s3 url
        :returns: S_OK(key) / S_ERROR
        """

        res = pfnparse(url, srmSpecific=False)
        if not res["OK"]:
            return res

        splitURL = res["Value"]

        # The path originally looks like '/bucket/lhcb/user/c/chaen
        # We remove the trailing slash, and get the relative path
        # of bucket/lhcb/user/c/chaen starting from bucket,
        # which gives you basically the LFN without trailing slash
        path = os.path.relpath(splitURL["Path"].lstrip("/"), start=self.bucketName)

        key = os.path.join(path, splitURL["FileName"])

        return S_OK(key)

    # @_extractKeyFromS3Path
    # def direct_exists(self, keys):
    #   """ Check if the keys exists on the storage

    #   :param self: self reference
    #   :param keys: list of keys
    #   :returns: Failed dictionary: {pfn : error message}
    #             Successful dictionary: {pfn : bool}
    #             S_ERROR in case of argument problems
    #   """

    #   successful = {}
    #   failed = {}

    #   # If we have a direct access, we can just do the request directly
    #   if self.directAccess:
    #     for key in keys:
    #       try:
    #         self.s3_client.head_object(Bucket=self.bucketName, Key=key)
    #         successful[key] = True
    #       except ClientError as exp:
    #         if exp.response['Error']['Code'] == '404':
    #           successful[key] = False
    #         else:
    #           failed[key] = repr(exp)
    #       except Exception as exp:
    #         failed[key] = repr(exp)
    #   else:
    #     # Otherwise, ask the gw for a presigned URL,
    #     # and perform it with requests
    #     for key in keys:
    #       try:
    #         res = self.S3GatewayClient.createPresignedUrl(self.name, 'head_object', key)
    #         if not res['OK']:
    #           failed[key] = res['Message']
    #           continue
    #         presignedURL = res['Value']
    #         response = requests.get(presignedURL)
    #         if response.status_code == 200:
    #           successful[key] = True
    #         elif response.status_code == 404:  # not found
    #           successful[key] = False
    #         else:
    #           failed[key] = response.reason
    #       except Exception as e:
    #         failed[key] = repr(e)

    #   resDict = {'Failed': failed, 'Successful': successful}
    #   return S_OK(resDict)

    def exists(self, urls):
        """Check if the urls exists on the storage

        :param urls: list of URLs
        :returns: Failed dictionary: {url : error message}
                  Successful dictionary: {url : bool}
                  S_ERROR in case of argument problems
        """

        if self.directAccess:
            return self._direct_exists(urls)

        return self._presigned_exists(urls)

    @_extractKeyFromS3Path
    def _direct_exists(self, urls):
        """Check if the files exists on the storage

        :param urls: list of urls
        :returns: Failed dictionary: {pfn : error message}
                  Successful dictionary: {pfn : bool}
                  S_ERROR in case of argument problems
        """

        successful = {}
        failed = {}

        # the @_extractKeyFromS3Path transformed URL into keys
        keys = urls

        for key in keys:
            try:
                self.s3_client.head_object(Bucket=self.bucketName, Key=key)
                successful[key] = True
            except ClientError as exp:
                if exp.response["Error"]["Code"] == "404":
                    successful[key] = False
                else:
                    failed[key] = repr(exp)
            except Exception as exp:
                failed[key] = repr(exp)

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def _presigned_exists(self, urls):
        """Check if the URLs exists on the storage

        :param urls: list of urls
        :returns: Failed dictionary: {pfn : error message}
                  Successful dictionary: {pfn : bool}
                  S_ERROR in case of argument problems
        """

        successful = {}
        failed = {}

        res = self.S3GatewayClient.createPresignedUrl(self.name, "head_object", urls)
        if not res["OK"]:
            return res

        failed.update(res["Value"]["Failed"])
        presignedURLs = res["Value"]["Successful"]

        # Otherwise, ask the gw for a presigned URL,
        # and perform it with requests
        for url, presignedURL in presignedURLs.items():
            try:
                response = requests.get(presignedURL)
                if response.status_code == 200:
                    successful[url] = True
                elif response.status_code == 404:  # not found
                    successful[url] = False
                else:
                    failed[url] = response.reason
            except Exception as e:
                failed[url] = repr(e)

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def isFile(self, urls):
        """Check if the urls provided are a file or not

        In practice, if the object exists, it is necessarily a file

        :param urls: list of urls to be checked
        :returns: * Failed dict: {path : error message}
                  * Successful dict: {path : bool}
                  * S_ERROR in case of argument problems

        """

        return self.exists(urls)

    def getFile(self, urls, localPath=False):
        """Make a local copy of the urls.

        :param  urls: list of urls on the storage
        :param localPath: destination folder. Default is from current directory
        :returns: * Successful dict: {path : size}
                  * Failed dict: {path : errorMessage}
                  * S_ERROR in case of argument problems
        """

        if self.directAccess:
            return self._direct_getFile(urls, localPath=localPath)
        return self._presigned_getFile(urls, localPath=localPath)

    @_extractKeyFromS3Path
    def _direct_getFile(self, urls, localPath=False):
        """Make a local copy of the keys.

        :param  urls: list of urls  on storage
        :param localPath: destination folder. Default is from current directory
        :returns: * Successful dict: {path : size}
                  * Failed dict: {path : errorMessage}
                  * S_ERROR in case of argument problems
        """

        log = LOG.getSubLogger("getFile")

        # the @_extractKeyFromS3Path transformed URL into keys
        keys = urls

        failed = {}
        successful = {}

        for src_key in keys:
            try:
                fileName = os.path.basename(src_key)
                dest_file = os.path.join(localPath if localPath else os.getcwd(), fileName)
                log.debug("Trying to download %s to %s" % (src_key, dest_file))

                self.s3_client.download_file(self.bucketName, src_key, dest_file)

                successful[src_key] = os.path.getsize(dest_file)
            except Exception as exp:
                failed[src_key] = repr(exp)

        return S_OK({"Failed": failed, "Successful": successful})

    def _presigned_getFile(self, urls, localPath=False):
        """Make a local copy of the files.

        :param  urls: list of urls  on storage
        :param localPath: destination folder. Default is from current directory
        :returns: * Successful dict: {path : size}
                  * Failed dict: {path : errorMessage}
                  * S_ERROR in case of argument problems
        """

        log = LOG.getSubLogger("getFile")

        failed = {}
        successful = {}

        res = self.S3GatewayClient.createPresignedUrl(self.name, "get_object", urls)
        if not res["OK"]:
            return res

        failed.update(res["Value"]["Failed"])

        presignedURLs = res["Value"]["Successful"]

        for src_url, presignedURL in presignedURLs.items():
            try:
                fileName = os.path.basename(src_url)
                dest_file = os.path.join(localPath if localPath else os.getcwd(), fileName)
                log.debug("Trying to download %s to %s" % (src_url, dest_file))

                # Stream download to save memory
                # https://requests.readthedocs.io/en/latest/user/advanced/#body-content-workflow
                with requests.get(presignedURL, stream=True) as r:
                    r.raise_for_status()
                    with open(dest_file, "wb") as f:
                        for chunk in r.iter_content():
                            if chunk:  # filter out keep-alive new chuncks
                                f.write(chunk)

                successful[src_url] = os.path.getsize(dest_file)
            except Exception as exp:
                failed[src_url] = repr(exp)

        return S_OK({"Failed": failed, "Successful": successful})

    def putFile(self, urls, sourceSize=0):
        """Upload a local file.

        ..warning:: no 3rd party copy possible

        :param urls: dictionary { urls : localFile }
        :param sourceSize: size of the file in byte. Mandatory for third party copy (WHY ???)
                             Also, this parameter makes it essentially a non bulk operation for
                             third party copy, unless all files have the same size...
        :returns: * Successful dict: { path : size }
                  * Failed dict: { path : error message }
                  * S_ERROR in case of argument problems
        """
        if self.directAccess:
            return self._direct_putFile(urls, sourceSize=sourceSize)
        return self._presigned_putFile(urls, sourceSize=sourceSize)

    @_extractKeyFromS3Path
    def _direct_putFile(self, urls, sourceSize=0):
        """Upload a local file.

        ..warning:: no 3rd party copy possible

        :param urls: dictionary { urls : localFile }
        :param sourceSize: size of the file in byte. Mandatory for third party copy (WHY ???)
                             Also, this parameter makes it essentially a non bulk operation for
                             third party copy, unless all files have the same size...
        :returns: * Successful dict: { path : size }
                  * Failed dict: { path : error message }
                  * S_ERROR in case of argument problems
        """

        log = LOG.getSubLogger("putFile")

        # the @_extractKeyFromS3Path transformed URL into keys
        keys = urls

        failed = {}
        successful = {}

        for dest_key, src_file in keys.items():
            try:
                cks = fileAdler(src_file)
                if not cks:
                    log.warn("Cannot get ADLER32 checksum for %s" % src_file)

                with open(src_file, "rb") as src_fd:
                    self.s3_client.put_object(
                        Body=src_fd, Bucket=self.bucketName, Key=dest_key, Metadata={"Checksum": cks}
                    )

                successful[dest_key] = os.path.getsize(src_file)

            except Exception as e:
                failed[dest_key] = repr(e)

        return S_OK({"Failed": failed, "Successful": successful})

    def _presigned_putFile(self, urls, sourceSize=0):
        """Upload a local file.

        ..warning:: no 3rd party copy possible

        :param urls: dictionary { urls : localFile }
        :param sourceSize: size of the file in byte. Mandatory for third party copy (WHY ???)
                             Also, this parameter makes it essentially a non bulk operation for
                             third party copy, unless all files have the same size...
        :returns: * Successful dict: { path : size }
                  * Failed dict: { path : error message }
                  * S_ERROR in case of argument problems
        """

        log = LOG.getSubLogger("putFile")

        failed = {}
        successful = {}

        # Construct a dict <url:{x-amz-meta-checksum: adler32}>
        # it needs to be passed to createPresignedUrl
        urlAdlers = {url: {"x-amz-meta-checksum": fileAdler(src_file)} for url, src_file in urls.items()}

        res = self.S3GatewayClient.createPresignedUrl(self.name, "put_object", urlAdlers)
        if not res["OK"]:
            return res

        failed.update(res["Value"]["Failed"])

        # Contains <url: presignedResponse>
        presignedResponses = res["Value"]["Successful"]

        for dest_url, presignedResponse in presignedResponses.items():

            src_file = urls[dest_url]

            try:
                cks = fileAdler(src_file)
                if not cks:
                    log.warn("Cannot get ADLER32 checksum for %s" % src_file)

                presignedURL = presignedResponse["url"]
                presignedFields = presignedResponse["fields"]
                with open(src_file, "rb") as src_fd:
                    # files = {'file': (dest_key, src_fd)}
                    files = {"file": src_fd}
                    response = requests.post(presignedURL, data=presignedFields, files=files)

                    if not response.ok:
                        raise Exception(response.reason)

                successful[dest_url] = os.path.getsize(src_file)

            except Exception as e:
                failed[dest_url] = repr(e)

        return S_OK({"Failed": failed, "Successful": successful})

    def getFileMetadata(self, urls):
        """Get metadata associated to the file(s)

        :param  urls: list of urls on the storage
        :returns: * successful dict { path : metadata }
                  * failed dict { path : error message }
                  * S_ERROR in case of argument problems
        """
        if self.directAccess:
            return self._direct_getFileMetadata(urls)
        return self._presigned_getFileMetadata(urls)

    @_extractKeyFromS3Path
    def _direct_getFileMetadata(self, urls):
        """Get metadata associated to the file(s)

        :param  urls: list of urls on the storage
        :returns: * successful dict { path : metadata }
                  * failed dict { path : error message }
                  * S_ERROR in case of argument problems
        """

        # the @_extractKeyFromS3Path transformed URL into keys
        keys = urls

        failed = {}
        successful = {}

        for key in keys:
            try:
                response = self.s3_client.head_object(Bucket=self.bucketName, Key=key)
                responseMetadata = response["ResponseMetadata"]["HTTPHeaders"]

                metadataDict = self._addCommonMetadata(responseMetadata)
                metadataDict["File"] = True
                metadataDict["Size"] = int(metadataDict["content-length"])
                metadataDict["Checksum"] = metadataDict.get("x-amz-meta-checksum", "")

                successful[key] = metadataDict
            except Exception as exp:
                failed[key] = repr(exp)

        return S_OK({"Failed": failed, "Successful": successful})

    def _presigned_getFileMetadata(self, urls):
        """Get metadata associated to the file(s)

        :param  urls: list of urls on the storage
        :returns: * successful dict { path : metadata }
                  * failed dict { path : error message }
                  * S_ERROR in case of argument problems
        """

        failed = {}
        successful = {}

        res = self.S3GatewayClient.createPresignedUrl(self.name, "head_object", urls)
        if not res["OK"]:
            return res

        failed.update(res["Value"]["Failed"])

        presignedURLs = res["Value"]["Successful"]

        for url, presignedURL in presignedURLs.items():
            try:
                response = requests.head(presignedURL)
                if not response.ok:
                    raise Exception(response.reason)

                # Although the interesting fields are the same as when doing the query directly
                # the case is not quite the same, so make it lower everywhere
                responseMetadata = {headerKey.lower(): headerVal for headerKey, headerVal in response.headers.items()}

                metadataDict = self._addCommonMetadata(responseMetadata)
                metadataDict["File"] = True
                metadataDict["Size"] = int(metadataDict["content-length"])
                metadataDict["Checksum"] = metadataDict.get("x-amz-meta-checksum", "")

                successful[url] = metadataDict
            except Exception as exp:
                failed[url] = repr(exp)

        return S_OK({"Failed": failed, "Successful": successful})

    def removeFile(self, urls):
        """Physically remove the file specified by keys

        A non existing file will be considered as successfully removed

        :param urls: list of urls on the storage
        :returns: * Successful dict {path : True}
                  * Failed dict {path : error message}
                  * S_ERROR in case of argument problems
        """

        if self.directAccess:
            return self._direct_removeFile(urls)
        return self._presigned_removeFile(urls)

    @_extractKeyFromS3Path
    def _direct_removeFile(self, urls):
        """Physically remove the file specified by keys

        A non existing file will be considered as successfully removed

        :param urls: list of urls on the storage
        :returns: * Successful dict {path : True}
                  * Failed dict {path : error message}
                  * S_ERROR in case of argument problems
        """

        failed = {}
        successful = {}

        # the @_extractKeyFromS3Path transformed URL into keys
        keys = urls

        for key in keys:
            try:
                self.s3_client.delete_object(Bucket=self.bucketName, Key=key)
                successful[key] = True
            except Exception as exp:
                failed[key] = repr(exp)

        return S_OK({"Failed": failed, "Successful": successful})

    def _presigned_removeFile(self, urls):
        """Physically remove the file specified by keys

        A non existing file will be considered as successfully removed

        :param urls: list of urls on the storage
        :returns: * Successful dict {path : True}
                  * Failed dict {path : error message}
                  * S_ERROR in case of argument problems
        """

        failed = {}
        successful = {}

        res = self.S3GatewayClient.createPresignedUrl(self.name, "delete_object", urls)
        if not res["OK"]:
            return res

        failed.update(res["Value"]["Failed"])

        presignedURLs = res["Value"]["Successful"]

        for url, presignedURL in presignedURLs.items():
            try:
                response = requests.delete(presignedURL)
                if not response.ok:
                    raise Exception(response.reason)

                successful[url] = True
            except Exception as exp:
                failed[url] = repr(exp)

        return S_OK({"Failed": failed, "Successful": successful})

    def getFileSize(self, urls):
        """Get the physical size of the given file

        :param urls: list of urls on the storage
        :returns: * Successful dict {path : size}
                  * Failed dict {path : error message }
                  * S_ERROR in case of argument problem
        """

        res = self.getFileMetadata(urls)
        if not res["OK"]:
            return res

        failed = res["Value"]["Failed"]
        successful = {url: metadata["Size"] for url, metadata in res["Value"]["Successful"].items()}

        return S_OK({"Successful": successful, "Failed": failed})

    #############################################################
    #
    # These are the methods for directory manipulation
    #

    def createDirectory(self, urls):
        """Create directory on the storage.
            S3 does not have such a concept, but we return OK for everything

        :param urls: list of urls to be created on the storage
        :returns: Always Successful dict {path : True }
        """

        return S_OK({"Failed": {}, "Successful": {url: True for url in urls}})

    @staticmethod
    def notAvailable(*_args, **_kwargs):
        """Generic method for unavailable method on S3"""
        return S_ERROR("Functionality not available on S3")

    listDirectory = (
        isDirectory
    ) = getDirectory = removeDirectory = getDirectorySize = getDirectoryMetadata = putDirectory = notAvailable

    def getTransportURL(self, urls, protocols):
        """Get a transport URL for given urls
            If http/https is requested, the URLs will be valid for 24hours

        :param dict urls: s3 urls
        :param list protocols: a list of acceptable transport protocols in priority order.
                          In practice, besides 's3', it can only be:

                          * 'https' if secureConnection is True
                          * 'http' othewise

        :returns: succ/failed dict url with required protocol

        """

        res = super(S3Storage, self).getTransportURL(urls, protocols)
        # if the result is OK or the error different than errno.EPROTONOSUPPORT
        # we just return
        if not cmpError(res, errno.EPROTONOSUPPORT):
            return res

        # We support only http if it is an insecured connection and https if it is a secured connection
        if self.secureConnection and "https" not in protocols:
            return S_ERROR(errno.EPROTONOSUPPORT, "Only https protocol is supported")
        elif not self.secureConnection and "http" not in protocols:
            return S_ERROR(errno.EPROTONOSUPPORT, "Only http protocol is supported")

        # Make the presigned URLs valid for 24h
        if self.directAccess:
            return self.createPresignedUrl(urls, "get_object", expiration=60 * 60 * 24)

        return self.S3GatewayClient.createPresignedUrl(self.name, "get_object", urls, expiration=60 * 60 * 24)

    @_extractKeyFromS3Path
    def createPresignedUrl(self, urls, s3_method, expiration=3600):
        """Generate a presigned URL to share an S3 object

        :param urls: urls for which to generate a presigned URL. If s3_method is put_object, it must be a dict <url:Fields>
                      where fields are the metadata of the file
                      (see https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.generate_presigned_post) # pylint: disable=line-too-long # noqa
        :param s3_method: name of the method for which to generate a presigned URL
        :param expiration: Time in seconds for the presigned URL to remain valid
        :return: Presigned URL as string. If error, returns None.
        """

        # the decorator transformed the urls into keys
        keys = urls

        successful = {}
        failed = {}

        # Generate a presigned URL for the S3 object
        log = LOG.getSubLogger("createPresignedUrl")

        for key in keys:
            try:
                if s3_method != "put_object":
                    response = self.s3_client.generate_presigned_url(
                        ClientMethod=s3_method, Params={"Bucket": self.bucketName, "Key": key}, ExpiresIn=expiration
                    )
                else:
                    fields = keys.get(key)
                    if not isinstance(fields, dict):
                        fields = None
                    response = self.s3_client.generate_presigned_post(
                        self.bucketName, key, Fields=fields, ExpiresIn=expiration
                    )

                successful[key] = response
            except ClientError as e:
                log.debug(e)
                failed[key] = repr(e)

        # The response contains the presigned URL
        return S_OK({"Successful": successful, "Failed": failed})



""" Base Storage Class provides the base interface for all storage plug-ins

      exists()

These are the methods for manipulating files:
      isFile()
      getFile()
      putFile()
      removeFile()
      getFileMetadata()
      getFileSize()
      prestageFile()
      getTransportURL()

These are the methods for manipulating directories:
      isDirectory()
      getDirectory()
      putDirectory()
      createDirectory()
      removeDirectory()
      listDirectory()
      getDirectoryMetadata()
      getDirectorySize()

These are the methods for manipulating the client:
      changeDirectory()
      getCurrentDirectory()
      getName()
      getParameters()
      getCurrentURL()

These are the methods for getting information about the Storage:
      getOccupancy()

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import errno
import json
import os
import shutil
import tempfile

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.Pfn import pfnparse, pfnunparse
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Resources.Storage.Utilities import checkArgumentFormat


class StorageBase(object):
    """
    .. class:: StorageBase

    """

    PROTOCOL_PARAMETERS = ["Protocol", "Host", "Path", "Port", "SpaceToken", "WSUrl"]
    # Options to be prepended in the URL
    # keys are the name of the parameters in the CS
    # values are the name of the options as they appear in the URL
    DYNAMIC_OPTIONS = {}

    def __init__(self, name, parameterDict):

        self.name = name
        self.pluginName = ""
        self.protocolParameters = {}

        self.__updateParameters(parameterDict)

        # Keep the list of all parameters passed for constructions
        # Taken from the CS
        # In a further major release, this could be nerged together
        # with protocolParameters. There is no reason for it to
        # be so strict about the possible content.
        self._allProtocolParameters = parameterDict

        if hasattr(self, "_INPUT_PROTOCOLS"):
            self.protocolParameters["InputProtocols"] = getattr(self, "_INPUT_PROTOCOLS")
        else:
            self.protocolParameters["InputProtocols"] = [self.protocolParameters["Protocol"], "file"]

        if hasattr(self, "_OUTPUT_PROTOCOLS"):
            self.protocolParameters["OutputProtocols"] = getattr(self, "_OUTPUT_PROTOCOLS")
        else:
            self.protocolParameters["OutputProtocols"] = [self.protocolParameters["Protocol"]]

        self.basePath = parameterDict["Path"]
        self.cwd = self.basePath
        self.se = None
        self.isok = True

        # use True for backward compatibility
        self.srmSpecificParse = True

    def setStorageElement(self, se):
        self.se = se

    def setParameters(self, parameterDict):
        """Set standard parameters, method can be overriden in subclasses
        to process specific parameters
        """
        self.__updateParameters(parameterDict)

    def __updateParameters(self, parameterDict):
        """setParameters implementation method"""
        for item in self.PROTOCOL_PARAMETERS:
            self.protocolParameters[item] = parameterDict.get(item, "")

    def getParameters(self):
        """Get the parameters with which the storage was instantiated"""
        parameterDict = dict(self.protocolParameters)
        parameterDict["StorageName"] = self.name
        parameterDict["PluginName"] = self.pluginName
        parameterDict["URLBase"] = self.getURLBase().get("Value", "")
        parameterDict["Endpoint"] = self.getEndpoint().get("Value", "")

        return parameterDict

    def exists(self, *parms, **kws):
        """Check if the given path exists"""
        return S_ERROR("Storage.exists: implement me!")

    #############################################################
    #
    # These are the methods for file manipulation
    #

    def isFile(self, *parms, **kws):
        """Check if the given path exists and it is a file"""
        return S_ERROR("Storage.isFile: implement me!")

    def getFile(self, *parms, **kws):
        """Get a local copy of the file specified by its path"""
        return S_ERROR("Storage.getFile: implement me!")

    def putFile(self, *parms, **kws):
        """Put a copy of the local file to the current directory on the
        physical storage
        """
        return S_ERROR("Storage.putFile: implement me!")

    def removeFile(self, *parms, **kws):
        """Remove physically the file specified by its path"""
        return S_ERROR("Storage.removeFile: implement me!")

    def getFileMetadata(self, *parms, **kws):
        """Get metadata associated to the file"""
        return S_ERROR("Storage.getFileMetadata: implement me!")

    def getFileSize(self, *parms, **kws):
        """Get the physical size of the given file"""
        return S_ERROR("Storage.getFileSize: implement me!")

    def prestageFile(self, *parms, **kws):
        """Issue prestage request for file"""
        return S_ERROR("Storage.prestageFile: implement me!")

    def prestageFileStatus(self, *parms, **kws):
        """Obtain the status of the prestage request"""
        return S_ERROR("Storage.prestageFileStatus: implement me!")

    def pinFile(self, *parms, **kws):
        """Pin the file on the destination storage element"""
        return S_ERROR("Storage.pinFile: implement me!")

    def releaseFile(self, *parms, **kws):
        """Release the file on the destination storage element"""
        return S_ERROR("Storage.releaseFile: implement me!")

    #############################################################
    #
    # These are the methods for directory manipulation
    #

    def isDirectory(self, *parms, **kws):
        """Check if the given path exists and it is a directory"""
        return S_ERROR("Storage.isDirectory: implement me!")

    def getDirectory(self, *parms, **kws):
        """Get locally a directory from the physical storage together with all its
        files and subdirectories.
        """
        return S_ERROR("Storage.getDirectory: implement me!")

    def putDirectory(self, *parms, **kws):
        """Put a local directory to the physical storage together with all its
        files and subdirectories.
        """
        return S_ERROR("Storage.putDirectory: implement me!")

    def createDirectory(self, *parms, **kws):
        """Make a new directory on the physical storage"""
        return S_ERROR("Storage.createDirectory: implement me!")

    def removeDirectory(self, *parms, **kws):
        """Remove a directory on the physical storage together with all its files and
        subdirectories.
        """
        return S_ERROR("Storage.removeDirectory: implement me!")

    def listDirectory(self, *parms, **kws):
        """List the supplied path"""
        return S_ERROR("Storage.listDirectory: implement me!")

    def getDirectoryMetadata(self, *parms, **kws):
        """Get the metadata for the directory"""
        return S_ERROR("Storage.getDirectoryMetadata: implement me!")

    def getDirectorySize(self, *parms, **kws):
        """Get the size of the directory on the storage"""
        return S_ERROR("Storage.getDirectorySize: implement me!")

    #############################################################
    #
    # These are the methods for manipulating the client
    #

    def isOK(self):
        return self.isok

    def resetCurrentDirectory(self):
        """Reset the working directory to the base dir"""
        self.cwd = self.basePath

    def changeDirectory(self, directory):
        """Change the directory to the supplied directory"""
        if directory.startswith("/"):
            self.cwd = "%s/%s" % (self.basePath, directory)
        else:
            self.cwd = "%s/%s" % (self.cwd, directory)

    def getCurrentDirectory(self):
        """Get the current directory"""
        return self.cwd

    def getCurrentURL(self, fileName):
        """Obtain the current file URL from the current working directory and the filename

        :param self: self reference
        :param str fileName: path on storage
        """
        urlDict = dict(self.protocolParameters)
        if not fileName.startswith("/"):
            # Relative path is given
            urlDict["Path"] = self.cwd
        result = pfnunparse(urlDict, srmSpecific=self.srmSpecificParse)
        if not result["OK"]:
            return result
        cwdUrl = result["Value"]
        fullUrl = "%s%s" % (cwdUrl, fileName)
        return S_OK(fullUrl)

    def getName(self):
        """The name with which the storage was instantiated"""
        return self.name

    def getURLBase(self, withWSUrl=False):
        """This will get the URL base. This is then appended with the LFN in DIRAC convention.

        :param self: self reference
        :param bool withWSUrl: flag to include Web Service part of the url
        :returns: URL
        """
        urlDict = dict(self.protocolParameters)
        if not withWSUrl:
            urlDict["WSUrl"] = ""
        return pfnunparse(urlDict, srmSpecific=self.srmSpecificParse)

    def getEndpoint(self):
        """This will get endpoint of the storage. It basically is the same as :py:meth:`getURLBase`
            but without the basePath

        :returns: 'proto://hostname<:port>'

        """
        urlDict = dict(self.protocolParameters)
        # We remove the basePath
        urlDict["Path"] = ""
        return pfnunparse(urlDict, srmSpecific=self.srmSpecificParse)

    def isURL(self, path):
        """Guess if the path looks like a URL

        :param self: self reference
        :param string path: input file LFN or URL
        :returns boolean: True if URL, False otherwise
        """
        if self.basePath and path.startswith(self.basePath):
            return S_OK(True)

        result = pfnparse(path, srmSpecific=self.srmSpecificParse)
        if not result["OK"]:
            return result

        if len(result["Value"]["Protocol"]) != 0:
            return S_OK(True)

        if result["Value"]["Path"].startswith(self.basePath):
            return S_OK(True)

        return S_OK(False)

    def getTransportURL(self, pathDict, protocols):
        """Get a transport URL for a given URL. For a simple storage plugin
        it is just returning input URL if the plugin protocol is one of the
        requested protocols

        :param dict pathDict: URL obtained from File Catalog or constructed according
                        to convention
        :param protocols: a list of acceptable transport protocols in priority order
        :type protocols: `python:list`
        """
        res = checkArgumentFormat(pathDict)
        if not res["OK"]:
            return res
        urls = res["Value"]
        successful = {}
        failed = {}

        if protocols and not self.protocolParameters["Protocol"] in protocols:
            return S_ERROR(errno.EPROTONOSUPPORT, "No native protocol requested")

        for url in urls:
            successful[url] = url

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def constructURLFromLFN(self, lfn, withWSUrl=False):
        """Construct URL from the given LFN according to the VO convention for the
        primary protocol of the storage plagin

        :param str lfn: file LFN
        :param boolean withWSUrl: flag to include the web service part into the resulting URL
        :return result: result['Value'] - resulting URL
        """

        # Check the LFN convention:
        # 1. LFN must start with the VO name as the top level directory
        # 2. VO name must not appear as any subdirectory or file name
        lfnSplitList = lfn.split("/")
        voLFN = lfnSplitList[1]
        # TODO comparison to Sandbox below is for backward compatibility, should
        # be removed in the next release
        if voLFN != self.se.vo and voLFN != "SandBox" and voLFN != "Sandbox":
            return S_ERROR("LFN (%s) path must start with VO name (%s)" % (lfn, self.se.vo))

        urlDict = dict(self.protocolParameters)
        urlDict["Options"] = "&".join(
            "%s=%s" % (optionName, urlDict[paramName])
            for paramName, optionName in self.DYNAMIC_OPTIONS.items()
            if urlDict.get(paramName)
        )
        if not withWSUrl:
            urlDict["WSUrl"] = ""
        urlDict["FileName"] = lfn.lstrip("/")

        return pfnunparse(urlDict, srmSpecific=self.srmSpecificParse)

    def updateURL(self, url, withWSUrl=False):
        """Update the URL according to the current SE parameters"""
        result = pfnparse(url, srmSpecific=self.srmSpecificParse)
        if not result["OK"]:
            return result
        urlDict = result["Value"]

        urlDict["Protocol"] = self.protocolParameters["Protocol"]
        urlDict["Host"] = self.protocolParameters["Host"]
        urlDict["Port"] = self.protocolParameters["Port"]
        urlDict["WSUrl"] = ""
        if withWSUrl:
            urlDict["WSUrl"] = self.protocolParameters["WSUrl"]

        return pfnunparse(urlDict, srmSpecific=self.srmSpecificParse)

    def isNativeURL(self, url):
        """Check if URL :url: is valid for :self.protocol:

        :param self: self reference
        :param str url: URL
        """
        res = pfnparse(url, srmSpecific=self.srmSpecificParse)
        if not res["OK"]:
            return res
        urlDict = res["Value"]
        return S_OK(urlDict["Protocol"] == self.protocolParameters["Protocol"])

    @staticmethod
    def _addCommonMetadata(metadataDict):
        """To make the output of getFileMetadata uniform throughout the protocols
        this returns a minimum set of metadata with default value,
        that are then complemented with the protocol specific metadata

        :param metadataDict: specific metadata of the protocol

        :returns: dictionnary with all the metadata (specific and basic)
        """
        commonMetadata = {
            "Checksum": "",
            "Directory": False,
            "File": False,
            "Mode": 0o000,
            "Size": 0,
            "Accessible": True,
        }

        commonMetadata.update(metadataDict)

        return commonMetadata

    def _isInputURL(self, url):
        """Check if the given url can be taken as input

        :param self: self reference
        :param str url: URL
        """
        res = pfnparse(url)
        if not res["OK"]:
            return res
        urlDict = res["Value"]

        # Special case of 'file' protocol which can be just a URL
        if not urlDict["Protocol"] and "file" in self.protocolParameters["InputProtocols"]:
            return S_OK(True)

        return S_OK(urlDict["Protocol"] == self.protocolParameters["Protocol"])

    #############################################################
    #
    # These are the methods for getting information about the Storage element:
    #

    def getOccupancy(self, **kwargs):
        """Get the StorageElement occupancy info in MB.

        This generic implementation download a json file supposed to contain the necessary info.

        :param occupancyLFN: (mandatory named argument) LFN of the json file.

        :returns: S_OK/S_ERROR dictionary. The S_OK value should contain a dictionary with Total and Free space in MB
        """

        # Build the URL for the occupancyLFN:
        occupancyLFN = kwargs["occupancyLFN"]
        res = self.constructURLFromLFN(occupancyLFN)
        if not res["OK"]:
            return res
        occupancyURL = res["Value"]

        try:

            # download the file locally
            tmpDirName = tempfile.mkdtemp()
            res = returnSingleResult(self.getFile(occupancyURL, localPath=tmpDirName))
            if not res["OK"]:
                return res

            filePath = os.path.join(tmpDirName, os.path.basename(occupancyLFN))

            # Read its json content
            with open(filePath, "r") as occupancyFile:
                return S_OK(json.load(occupancyFile))

        except Exception as e:
            return S_ERROR(repr(e))

        finally:
            # Clean the temporary dir
            shutil.rmtree(tmpDirName)



"""
This is the interface to DIRAC PilotAgentsDB.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import shutil
from DIRAC import S_OK, S_ERROR, gLogger
import DIRAC.Core.Utilities.Time as Time

from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getUsernameForDN, getDNForUsername
from DIRAC.WorkloadManagementSystem.Client import PilotStatus
from DIRAC.WorkloadManagementSystem.Service.WMSUtilities import (
    getPilotCE,
    getPilotProxy,
    getPilotRef,
    killPilotsInQueues,
)


class PilotManagerHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """Initialization of DB objects"""

        try:
            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.PilotAgentsDB", "PilotAgentsDB")
            if not result["OK"]:
                return result
            cls.pilotAgentsDB = result["Value"]()

        except RuntimeError as excp:
            return S_ERROR("Can't connect to DB: %s" % excp)

        cls.pilotsLoggingDB = None
        enablePilotsLogging = Operations().getValue("/Services/JobMonitoring/usePilotsLoggingFlag", False)
        if enablePilotsLogging:
            try:
                result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.PilotsLoggingDB", "PilotsLoggingDB")
                if not result["OK"]:
                    return result
                cls.pilotsLoggingDB = result["Value"]()
            except RuntimeError as excp:
                return S_ERROR("Can't connect to DB: %s" % excp)

        return S_OK()

    ##############################################################################
    types_getCurrentPilotCounters = [dict]

    @classmethod
    def export_getCurrentPilotCounters(cls, attrDict={}):
        """Get pilot counters per Status with attrDict selection. Final statuses are given for
        the last day.
        """

        result = cls.pilotAgentsDB.getCounters("PilotAgents", ["Status"], attrDict, timeStamp="LastUpdateTime")
        if not result["OK"]:
            return result
        last_update = Time.dateTime() - Time.day
        resultDay = cls.pilotAgentsDB.getCounters(
            "PilotAgents", ["Status"], attrDict, newer=last_update, timeStamp="LastUpdateTime"
        )
        if not resultDay["OK"]:
            return resultDay

        resultDict = {}
        for statusDict, count in result["Value"]:
            status = statusDict["Status"]
            resultDict[status] = count
            if status in PilotStatus.PILOT_FINAL_STATES:
                resultDict[status] = 0
                for statusDayDict, ccount in resultDay["Value"]:
                    if status == statusDayDict["Status"]:
                        resultDict[status] = ccount
                    break

        return S_OK(resultDict)

    ##########################################################################################
    types_addPilotTQReference = [list, six.integer_types, six.string_types, six.string_types]

    @classmethod
    def export_addPilotTQReference(
        cls, pilotRef, taskQueueID, ownerDN, ownerGroup, broker="Unknown", gridType="DIRAC", pilotStampDict={}
    ):
        """Add a new pilot job reference"""
        return cls.pilotAgentsDB.addPilotTQReference(
            pilotRef, taskQueueID, ownerDN, ownerGroup, broker, gridType, pilotStampDict
        )

    ##############################################################################
    types_getPilotOutput = [six.string_types]

    def export_getPilotOutput(self, pilotReference):
        """Get the pilot job standard output and standard error files for the Grid
        job reference
        """
        result = self.pilotAgentsDB.getPilotInfo(pilotReference)
        if not result["OK"]:
            gLogger.error("Failed to get info for pilot", result["Message"])
            return S_ERROR("Failed to get info for pilot")
        if not result["Value"]:
            gLogger.warn("The pilot info is empty", pilotReference)
            return S_ERROR("Pilot info is empty")

        pilotDict = result["Value"][pilotReference]

        owner = pilotDict["OwnerDN"]
        group = pilotDict["OwnerGroup"]

        # FIXME: What if the OutputSandBox is not StdOut and StdErr, what do we do with other files?
        result = self.pilotAgentsDB.getPilotOutput(pilotReference)
        if result["OK"]:
            stdout = result["Value"]["StdOut"]
            error = result["Value"]["StdErr"]
            if stdout or error:
                resultDict = {}
                resultDict["StdOut"] = stdout
                resultDict["StdErr"] = error
                resultDict["OwnerDN"] = owner
                resultDict["OwnerGroup"] = group
                resultDict["FileList"] = []
                return S_OK(resultDict)
            else:
                gLogger.warn("Empty pilot output found", "for %s" % pilotReference)

        result = getPilotCE(pilotDict)
        if not result["OK"]:
            return result

        ce = result["Value"]
        if not hasattr(ce, "getJobOutput"):
            return S_ERROR("Pilot output not available for %s CEs" % pilotDict["GridType"])

        result = getPilotProxy(pilotDict)
        if not result["OK"]:
            return result

        proxy = result["Value"]
        ce.setProxy(proxy)

        result = getPilotRef(pilotReference, pilotDict)
        if not result["OK"]:
            return result
        pRef = result["Value"]

        result = ce.getJobOutput(pRef)
        if not result["OK"]:
            shutil.rmtree(ce.ceParameters["WorkingDirectory"])
            return result
        stdout, error = result["Value"]
        if stdout:
            result = self.pilotAgentsDB.storePilotOutput(pilotReference, stdout, error)
            if not result["OK"]:
                gLogger.error("Failed to store pilot output:", result["Message"])

        resultDict = {}
        resultDict["StdOut"] = stdout
        resultDict["StdErr"] = error
        resultDict["OwnerDN"] = owner
        resultDict["OwnerGroup"] = group
        resultDict["FileList"] = []
        shutil.rmtree(ce.ceParameters["WorkingDirectory"])
        return S_OK(resultDict)

    ##############################################################################
    types_getPilotInfo = [(list,) + six.string_types]

    @classmethod
    def export_getPilotInfo(cls, pilotReference):
        """Get the info about a given pilot job reference"""
        return cls.pilotAgentsDB.getPilotInfo(pilotReference)

    ##############################################################################
    types_selectPilots = [dict]

    @classmethod
    def export_selectPilots(cls, condDict):
        """Select pilots given the selection conditions"""
        return cls.pilotAgentsDB.selectPilots(condDict)

    ##############################################################################
    types_storePilotOutput = [six.string_types, six.string_types, six.string_types]

    @classmethod
    def export_storePilotOutput(cls, pilotReference, output, error):
        """Store the pilot output and error"""
        return cls.pilotAgentsDB.storePilotOutput(pilotReference, output, error)

    ##############################################################################
    types_getPilotLoggingInfo = [six.string_types]

    @classmethod
    def export_getPilotLoggingInfo(cls, pilotReference):
        """Get the pilot logging info for the Grid job reference"""
        result = cls.pilotAgentsDB.getPilotInfo(pilotReference)
        if not result["OK"]:
            gLogger.error("Failed to get info for pilot", result["Message"])
            return S_ERROR("Failed to get info for pilot")
        if not result["Value"]:
            gLogger.warn("The pilot info is empty", pilotReference)
            return S_ERROR("Pilot info is empty")

        pilotDict = result["Value"][pilotReference]
        result = getPilotCE(pilotDict)
        if not result["OK"]:
            return result

        ce = result["Value"]
        if not hasattr(ce, "getJobLog"):
            gLogger.info("Pilot logging not available for", "%s CEs" % pilotDict["GridType"])
            return S_ERROR("Pilot logging not available for %s CEs" % pilotDict["GridType"])

        result = getPilotProxy(pilotDict)
        if not result["OK"]:
            return result

        proxy = result["Value"]
        ce.setProxy(proxy)

        result = getPilotRef(pilotReference, pilotDict)
        if not result["OK"]:
            return result
        pRef = result["Value"]

        result = ce.getJobLog(pRef)
        if not result["OK"]:
            shutil.rmtree(ce.ceParameters["WorkingDirectory"])
            return result
        loggingInfo = result["Value"]
        shutil.rmtree(ce.ceParameters["WorkingDirectory"])

        return S_OK(loggingInfo)

    ##############################################################################
    types_getPilotSummary = []

    @classmethod
    def export_getPilotSummary(cls, startdate="", enddate=""):
        """Get summary of the status of the LCG Pilot Jobs"""

        return cls.pilotAgentsDB.getPilotSummary(startdate, enddate)

    ##############################################################################
    types_getPilotMonitorWeb = [dict, list, six.integer_types, list(six.integer_types)]

    @classmethod
    def export_getPilotMonitorWeb(cls, selectDict, sortList, startItem, maxItems):
        """Get the summary of the pilot information for a given page in the
        pilot monitor in a generic format
        """

        return cls.pilotAgentsDB.getPilotMonitorWeb(selectDict, sortList, startItem, maxItems)

    ##############################################################################
    types_getPilotMonitorSelectors = []

    @classmethod
    def export_getPilotMonitorSelectors(cls):
        """Get all the distinct selector values for the Pilot Monitor web portal page"""

        return cls.pilotAgentsDB.getPilotMonitorSelectors()

    ##############################################################################
    types_getPilotSummaryWeb = [dict, list, six.integer_types, list(six.integer_types)]

    @classmethod
    def export_getPilotSummaryWeb(cls, selectDict, sortList, startItem, maxItems):
        """Get the summary of the pilot information for a given page in the
        pilot monitor in a generic format
        """

        return cls.pilotAgentsDB.getPilotSummaryWeb(selectDict, sortList, startItem, maxItems)

    ##############################################################################
    types_getGroupedPilotSummary = [dict, list]

    @classmethod
    def export_getGroupedPilotSummary(cls, selectDict, columnList):
        """
        Get pilot summary showing grouped by columns in columnList, all pilot states
        and pilot efficiencies in a single row.

        :param selectDict: additional arguments to SELECT clause
        :param columnList: a list of columns to GROUP BY (less status column)
        :return: a dictionary containing column names and data records
        """
        return cls.pilotAgentsDB.getGroupedPilotSummary(selectDict, columnList)

    ##############################################################################
    types_getPilots = [six.string_types + six.integer_types]

    @classmethod
    def export_getPilots(cls, jobID):
        """Get pilot references and their states for :
        - those pilots submitted for the TQ where job is sitting
        - (or) the pilots executing/having executed the Job
        """

        pilots = []
        result = cls.pilotAgentsDB.getPilotsForJobID(int(jobID))
        if not result["OK"]:
            if result["Message"].find("not found") == -1:
                return S_ERROR("Failed to get pilot: " + result["Message"])
        else:
            pilots += result["Value"]
        if not pilots:
            # Pilots were not found try to look in the Task Queue
            taskQueueID = 0
            try:
                result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.TaskQueueDB", "TaskQueueDB")
                if not result["OK"]:
                    return result
                tqDB = result["Value"]()
            except RuntimeError as excp:
                return S_ERROR("Can't connect to DB: %s" % excp)
            result = tqDB.getTaskQueueForJob(int(jobID))
            if result["OK"] and result["Value"]:
                taskQueueID = result["Value"]
            if taskQueueID:
                result = cls.pilotAgentsDB.getPilotsForTaskQueue(taskQueueID, limit=10)
                if not result["OK"]:
                    return S_ERROR("Failed to get pilot: " + result["Message"])
                pilots += result["Value"]

        if not pilots:
            return S_ERROR("Failed to get pilot for Job %d" % int(jobID))

        return cls.pilotAgentsDB.getPilotInfo(pilotID=pilots)

    ##############################################################################
    types_killPilot = [six.string_types + (list,)]

    @classmethod
    def export_killPilot(cls, pilotRefList):
        """Kill the specified pilots"""
        # Make a list if it is not yet
        pilotRefs = list(pilotRefList)
        if isinstance(pilotRefList, six.string_types):
            pilotRefs = [pilotRefList]

        # Regroup pilots per site and per owner
        pilotRefDict = {}
        for pilotReference in pilotRefs:
            result = cls.pilotAgentsDB.getPilotInfo(pilotReference)
            if not result["OK"] or not result["Value"]:
                return S_ERROR("Failed to get info for pilot " + pilotReference)

            pilotDict = result["Value"][pilotReference]
            owner = pilotDict["OwnerDN"]
            group = pilotDict["OwnerGroup"]
            queue = "@@@".join([owner, group, pilotDict["GridSite"], pilotDict["DestinationSite"], pilotDict["Queue"]])
            gridType = pilotDict["GridType"]
            pilotRefDict.setdefault(queue, {})
            pilotRefDict[queue].setdefault("PilotList", [])
            pilotRefDict[queue]["PilotList"].append(pilotReference)
            pilotRefDict[queue]["GridType"] = gridType

        failed = killPilotsInQueues(pilotRefDict)

        if failed:
            return S_ERROR("Failed to kill at least some pilots")

        return S_OK()

    ##############################################################################
    types_setJobForPilot = [six.string_types + six.integer_types, six.string_types]

    @classmethod
    def export_setJobForPilot(cls, jobID, pilotRef, destination=None):
        """Report the DIRAC job ID which is executed by the given pilot job"""

        result = cls.pilotAgentsDB.setJobForPilot(int(jobID), pilotRef)
        if not result["OK"]:
            return result
        result = cls.pilotAgentsDB.setCurrentJobID(pilotRef, int(jobID))
        if not result["OK"]:
            return result
        if destination:
            result = cls.pilotAgentsDB.setPilotDestinationSite(pilotRef, destination)

        return result

    ##########################################################################################
    types_setPilotBenchmark = [six.string_types, float]

    @classmethod
    def export_setPilotBenchmark(cls, pilotRef, mark):
        """Set the pilot agent benchmark"""
        return cls.pilotAgentsDB.setPilotBenchmark(pilotRef, mark)

    ##########################################################################################
    types_setAccountingFlag = [six.string_types]

    @classmethod
    def export_setAccountingFlag(cls, pilotRef, mark="True"):
        """Set the pilot AccountingSent flag"""
        return cls.pilotAgentsDB.setAccountingFlag(pilotRef, mark)

    ##########################################################################################
    types_setPilotStatus = [six.string_types, six.string_types]

    @classmethod
    def export_setPilotStatus(cls, pilotRef, status, destination=None, reason=None, gridSite=None, queue=None):
        """Set the pilot agent status"""

        return cls.pilotAgentsDB.setPilotStatus(
            pilotRef, status, destination=destination, statusReason=reason, gridSite=gridSite, queue=queue
        )

    ##########################################################################################
    types_countPilots = [dict]

    @classmethod
    def export_countPilots(cls, condDict, older=None, newer=None, timeStamp="SubmissionTime"):
        """Set the pilot agent status"""

        return cls.pilotAgentsDB.countPilots(condDict, older, newer, timeStamp)

    ##########################################################################################
    types_getCounters = [six.string_types, list, dict]

    @classmethod
    def export_getCounters(cls, table, keys, condDict, newer=None, timeStamp="SubmissionTime"):
        """Set the pilot agent status"""

        return cls.pilotAgentsDB.getCounters(table, keys, condDict, newer=newer, timeStamp=timeStamp)

    ##############################################################################
    types_getPilotStatistics = [six.string_types, dict]

    @classmethod
    def export_getPilotStatistics(cls, attribute, selectDict):
        """Get pilot statistics distribution per attribute value with a given selection"""

        startDate = selectDict.get("FromDate", None)
        if startDate:
            del selectDict["FromDate"]

        if startDate is None:
            startDate = selectDict.get("LastUpdate", None)
            if startDate:
                del selectDict["LastUpdate"]
        endDate = selectDict.get("ToDate", None)
        if endDate:
            del selectDict["ToDate"]

        # Owner attribute is not part of PilotAgentsDB
        # It has to be converted into a OwnerDN
        owners = selectDict.get("Owner")
        if owners:
            ownerDNs = []
            for owner in owners:
                result = getDNForUsername(owner)
                if not result["OK"]:
                    return result
                ownerDNs.append(result["Value"])

            selectDict["OwnerDN"] = ownerDNs
            del selectDict["Owner"]

        result = cls.pilotAgentsDB.getCounters(
            "PilotAgents", [attribute], selectDict, newer=startDate, older=endDate, timeStamp="LastUpdateTime"
        )
        statistics = {}
        if result["OK"]:
            for status, count in result["Value"]:
                if "OwnerDN" in status:
                    userName = getUsernameForDN(status["OwnerDN"])
                    if userName["OK"]:
                        status["OwnerDN"] = userName["Value"]
                    statistics[status["OwnerDN"]] = count
                else:
                    statistics[status[attribute]] = count

        return S_OK(statistics)

    ##############################################################################
    types_deletePilots = [(list,) + six.integer_types + six.string_types]

    @classmethod
    def export_deletePilots(cls, pilotIDs):

        if isinstance(pilotIDs, six.string_types):
            return cls.pilotAgentsDB.deletePilot(pilotIDs)

        if isinstance(pilotIDs, six.integer_types):
            pilotIDs = [
                pilotIDs,
            ]

        result = cls.pilotAgentsDB.deletePilots(pilotIDs)
        if not result["OK"]:
            return result
        if cls.pilotsLoggingDB:
            pilotIDs = result["Value"]
            pilots = cls.pilotAgentsDB.getPilotInfo(pilotID=pilotIDs)
            if not pilots["OK"]:
                return pilots
            pilotRefs = []
            for pilot in pilots:
                pilotRefs.append(pilot["PilotJobReference"])
            result = cls.pilotsLoggingDB.deletePilotsLogging(pilotRefs)
            if not result["OK"]:
                return result

        return S_OK()

    ##############################################################################
    types_clearPilots = [six.integer_types, six.integer_types]

    @classmethod
    def export_clearPilots(cls, interval=30, aborted_interval=7):

        result = cls.pilotAgentsDB.clearPilots(interval, aborted_interval)
        if not result["OK"]:
            return result
        if cls.pilotsLoggingDB:
            pilotIDs = result["Value"]
            pilots = cls.pilotAgentsDB.getPilotInfo(pilotID=pilotIDs)
            if not pilots["OK"]:
                return pilots
            pilotRefs = []
            for pilot in pilots:
                pilotRefs.append(pilot["PilotJobReference"])
            result = cls.pilotsLoggingDB.deletePilotsLogging(pilotRefs)
            if not result["OK"]:
                return result

        return S_OK()



""" Test class for ElementInspectorAgent
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# imports
from mock import MagicMock

# DIRAC Components
from DIRAC.ResourceStatusSystem.Agent.ElementInspectorAgent import ElementInspectorAgent
from DIRAC import gLogger

gLogger.setLevel("DEBUG")

# Mock Objects
mockAM = MagicMock()
mockNone = MagicMock()
mockNone.return_value = None
mockSM = MagicMock()

elemDict = {
    "status": "status",
    "name": "site",
    "site": "site",
    "vO": "all",
    "element": "Site",
    "statusType": "all",
    "elementType": "Site",
}


def test__execute(mocker):
    """Testing JobCleaningAgent()._getAllowedJobTypes()"""

    mocker.patch("DIRAC.ResourceStatusSystem.Agent.ElementInspectorAgent.AgentModule.__init__")
    mocker.patch("DIRAC.ResourceStatusSystem.Agent.ElementInspectorAgent.PEP")
    mocker.patch(
        "DIRAC.ResourceStatusSystem.Agent.ElementInspectorAgent.AgentModule._AgentModule__moduleProperties",
        side_effect=lambda x, y=None: y,
        create=True,
    )

    elementInspectorAgent = ElementInspectorAgent()
    elementInspectorAgent.log = gLogger
    elementInspectorAgent.log.setLevel("DEBUG")
    elementInspectorAgent._AgentModule__configDefaults = mockAM
    elementInspectorAgent.initialize()

    result = elementInspectorAgent._execute(elemDict)

    assert result is None



""" Test case for DIRAC.Core.Utilities.File module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

##
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2011/01/17 14:01:18
# @brief Definition of FileTestCase class.

# imports
import os
from os.path import abspath
import re
import sys

from hypothesis import given
from hypothesis.strategies import floats

from pytest import mark

# sut
from DIRAC.Core.Utilities.File import (
    checkGuid,
    makeGuid,
    getSize,
    getMD5ForFiles,
    convertSizeUnits,
    SIZE_UNIT_CONVERSION,
)

parametrize = mark.parametrize


__RCSID__ = "$Id$"


def testCheckGuid():
    """checkGuid tests"""
    # empty string
    guid = ""
    assert checkGuid(guid) is False, "empty guid"

    # wrong length in a 1st field
    guid = "012345678-0123-0123-0123-0123456789AB"
    assert checkGuid(guid) is False, "wrong length in 1st field"
    guid = "0123456-0123-0123-0123-0123456789AB"
    assert checkGuid(guid) is False, "wrong length in 1st field"

    # wrong length in a 2nd field
    guid = "01234567-01234-0123-0123-0123456789AB"
    assert checkGuid(guid) is False, "wrong length in 2nd field"
    guid = "01234567-012-0123-0123-0123456789AB"
    assert checkGuid(guid) is False, "wrong length in 2nd field"

    # wrong length in a 3rd field
    guid = "01234567-0123-01234-0123-0123456789AB"
    assert checkGuid(guid) is False, "wrong length in 3rd field"
    guid = "01234567-0123-012-0123-0123456789AB"
    assert checkGuid(guid) is False, "wrong length in 3rd field"

    # wrong length in a 4th field
    guid = "01234567-0123-0123-01234-0123456789AB"
    assert checkGuid(guid) is False, "wrong length in 4th field"
    guid = "01234567-0123-0123-012-0123456789AB"
    assert checkGuid(guid) is False, "wrong length in 4th field"

    # wrong length in a 5th field
    guid = "01234567-0123-0123-0123-0123-0123456789ABC"
    assert checkGuid(guid) is False, "wrong length in 5th field"
    guid = "01234567-0123-0123-0123-0123-0123456789A"
    assert checkGuid(guid) is False, "wrong length in 5th field"

    # small caps
    guid = "01234567-9ABC-0DEF-0123-456789ABCDEF".lower()
    assert checkGuid(guid) is True, "small caps in guid, zut!"

    # wrong characters not in [0-9A-F]
    guid = "NEEDMORE-SPAM-SPAM-SPAM-SPAMWITHEGGS"
    assert checkGuid(guid) is True, "wrong set of characters, zut!"

    # normal operation
    guid = "01234567-9ABC-0DEF-0123-456789ABCDEF"
    assert checkGuid(guid) is True, "proper GUID"


def testMakeGuid():
    """makeGuid tests"""
    # no filename - fake guid produced
    assert checkGuid(makeGuid()) is True, "fake guid for inexisting file"
    # using this python file
    assert checkGuid(makeGuid(abspath(__file__))) is True, "guid for FileTestCase.py file"


def testGetSize():
    """getSize tests"""
    # non existing file
    assert getSize("/spam/eggs/eggs") == -1, "inexisting file"
    # file unreadable
    assert getSize("/root/.login") == -1, "unreadable file"


def testGetMD5ForFiles():
    """getMD5ForFiles tests"""

    filesList = [abspath(".") + os.sep + x for x in os.listdir(".")]
    md5sum = getMD5ForFiles(filesList)
    reMD5 = re.compile("^[0-9a-fA-F]+$")
    assert reMD5.match(md5sum) is not None
    # OK for python 2.7
    # self.assertRegexpMatches( md5sum, reMD5, "regexp doesn't match" )


@given(nb=floats(allow_nan=False, allow_infinity=False, min_value=1))
def test_convert_to_bigger_unit_floats(nb):
    """Make sure that converting to bigger unit gets the number smaller .
    Also tests that two steps are equal to two consecutive steps
    """
    toKB = convertSizeUnits(nb, "B", "kB")
    toMB = convertSizeUnits(nb, "B", "MB")
    fromkBtoMB = convertSizeUnits(toKB, "kB", "MB")

    assert toKB < nb
    assert toMB < toKB
    assert toMB == fromkBtoMB


def test_convert_error_to_maxint():
    """Make sure that on error we receive -sys.maxint"""
    assert convertSizeUnits("size", "B", "kB") == -sys.maxsize
    assert convertSizeUnits(0, "srcUnit", "kB") == -sys.maxsize
    assert convertSizeUnits(0, "B", "dstUnit") == -sys.maxsize


@given(nb=floats(allow_nan=False, allow_infinity=False, min_value=1))
@parametrize("srcUnit", SIZE_UNIT_CONVERSION)
@parametrize("dstUnit", SIZE_UNIT_CONVERSION)
def test_convert_loop(nb, srcUnit, dstUnit):
    """Make sure that converting a size back and forth preserves the number"""

    converted = convertSizeUnits(convertSizeUnits(nb, srcUnit, dstUnit), dstUnit, srcUnit)
    # We exclude the infinity case
    if converted != float("Inf"):
        assert converted == nb



#!/usr/bin/env python
"""
Remove the given file replica or a list of file replicas from the File Catalog
This script should be used with great care as it may leave dark data in the storage!
Use dirac-dms-remove-replicas instead
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os

from DIRAC import exit as dexit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC import gLogger


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(("LocalFile: Path to local file containing LFNs", "LFN:       Logical File Names"))
    Script.registerArgument(" SE:        Storage element")
    Script.parseCommandLine()

    from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

    allowUsers = Operations().getValue("DataManagement/AllowUserReplicaManagement", False)

    from DIRAC.Core.Security.ProxyInfo import getProxyInfo

    res = getProxyInfo()
    if not res["OK"]:
        gLogger.fatal("Can't get proxy info", res["Message"])
        dexit(1)
    properties = res["Value"].get("groupProperties", [])

    if not allowUsers:
        if "FileCatalogManagement" not in properties:
            gLogger.error("You need to use a proxy from a group with FileCatalogManagement")
            dexit(5)

    from DIRAC.DataManagementSystem.Client.DataManager import DataManager

    dm = DataManager()
    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    inputFileName, storageElementName = Script.getPositionalArgs(group=True)

    if os.path.exists(inputFileName):
        inputFile = open(inputFileName, "r")
        string = inputFile.read()
        lfns = [lfn.strip() for lfn in string.splitlines()]
        inputFile.close()
    else:
        lfns = [inputFileName]

    res = dm.removeReplicaFromCatalog(storageElementName, lfns)
    if not res["OK"]:
        print(res["Message"])
        dexit(0)
    for lfn in sorted(res["Value"]["Failed"]):
        message = res["Value"]["Failed"][lfn]
        print("Failed to remove %s replica of %s: %s" % (storageElementName, lfn, message))
    print("Successfully remove %d catalog replicas at %s" % (len(res["Value"]["Successful"]), storageElementName))


if __name__ == "__main__":
    main()



#!/usr/bin/env python

# it expects to be called with the the number of processors allowed to use

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys
from multiprocessing import Pool, current_process

##############################################################################
# copy from List.py for convenience


def getChunk(aList, chunkSize):
    """Generator yielding chunk from a list of a size chunkSize.

    :param aList: list to be splitted
    :type aList: python:list
    :param int chunkSize: lenght of one chunk
    :raise: StopIteration

    Usage:

    >>> for chunk in getChunk( aList, chunkSize=10):
          process( chunk )

    """
    for i in range(0, len(aList), chunkSize):
        yield aList[i : i + chunkSize]


def breakListIntoChunks(aList, chunkSize):
    """This function takes a list as input and breaks it into list of size 'chunkSize'.
       It returns a list of lists.

    :param aList: list of elements
    :type aList: python:list
    :param int chunkSize: len of a single chunk
    :return: list of lists of length of chunkSize
    :raise: RuntimeError if numberOfFilesInChunk is less than 1
    """
    if chunkSize < 1:
        raise RuntimeError("chunkSize cannot be less than 1")
    if isinstance(aList, (set, dict, tuple)):
        aList = list(aList)
    return [chunk for chunk in getChunk(aList, chunkSize)]


##############################################################################


def do_sum(l):
    print(current_process().name)
    return sum(l)


def main(nProc):
    p = Pool(nProc)
    r = list(range(1, 20000000))
    chunkSize = int(20000000 / nProc)
    rc = breakListIntoChunks(r, chunkSize)
    r = p.map(do_sum, rc[:nProc])
    print(r, sum(r))


if __name__ == "__main__":
    nProc = sys.argv[1]
    main(int(nProc))



""" JobManagerHandler is the implementation of the JobManager service
    in the DISET framework

    The following methods are available in the Service interface

    submitJob()
    rescheduleJob()
    deleteJob()
    killJob()

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.DISET.MessageClient import MessageClient
from DIRAC.Core.Utilities.DErrno import EWMSJDL, EWMSSUBM
from DIRAC.Core.Utilities.ClassAd.ClassAdLight import ClassAd
from DIRAC.Core.Utilities.JEncode import strToIntDict
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Utilities.ParametricJob import generateParametricJobs, getParameterVectorLength
from DIRAC.WorkloadManagementSystem.Service.JobPolicy import (
    JobPolicy,
    RIGHT_SUBMIT,
    RIGHT_RESCHEDULE,
    RIGHT_DELETE,
    RIGHT_KILL,
    RIGHT_RESET,
)

MAX_PARAMETRIC_JOBS = 20


class JobManagerHandlerMixin(object):
    """RequestHandler implementation of the JobManager"""

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """Initialization of DB objects and OptimizationMind"""
        try:
            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.JobDB", "JobDB")
            if not result["OK"]:
                return result
            cls.jobDB = result["Value"]()

            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.JobLoggingDB", "JobLoggingDB")
            if not result["OK"]:
                return result
            cls.jobLoggingDB = result["Value"]()

            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.TaskQueueDB", "TaskQueueDB")
            if not result["OK"]:
                return result
            cls.taskQueueDB = result["Value"]()

            result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.PilotAgentsDB", "PilotAgentsDB")
            if not result["OK"]:
                return result
            cls.pilotAgentsDB = result["Value"]()

        except RuntimeError as excp:
            return S_ERROR("Can't connect to DB: %s" % excp)

        cls.pilotsLoggingDB = None
        enablePilotsLogging = Operations().getValue("/Services/JobMonitoring/usePilotsLoggingFlag", False)
        if enablePilotsLogging:
            try:
                result = ObjectLoader().loadObject("WorkloadManagementSystem.DB.PilotsLoggingDB", "PilotsLoggingDB")
                if not result["OK"]:
                    return result
                cls.pilotsLoggingDB = result["Value"]()
            except RuntimeError as excp:
                return S_ERROR("Can't connect to DB: %s" % excp)

        cls.msgClient = MessageClient("WorkloadManagement/OptimizationMind")
        result = cls.msgClient.connect(JobManager=True)
        if not result["OK"]:
            cls.log.warn("Cannot connect to OptimizationMind!", result["Message"])
        return S_OK()

    def initializeRequest(self):
        credDict = self.getRemoteCredentials()
        self.ownerDN = credDict["DN"]
        self.ownerGroup = credDict["group"]
        self.userProperties = credDict["properties"]
        self.owner = credDict["username"]
        self.peerUsesLimitedProxy = credDict["isLimitedProxy"]
        self.maxParametricJobs = self.srv_getCSOption("MaxParametricJobs", MAX_PARAMETRIC_JOBS)
        self.jobPolicy = JobPolicy(self.ownerDN, self.ownerGroup, self.userProperties)
        self.jobPolicy.jobDB = self.jobDB
        return S_OK()

    def __sendJobsToOptimizationMind(self, jids):
        if not self.msgClient.connected:
            result = self.msgClient.connect(JobManager=True)
            if not result["OK"]:
                self.log.warn("Cannot connect to OptimizationMind!", result["Message"])
                return

        result = self.msgClient.createMessage("OptimizeJobs")
        if not result["OK"]:
            self.log.error("Cannot create Optimize message", result["Message"])
            return
        msgObj = result["Value"]
        msgObj.jids = list(sorted(jids))
        result = self.msgClient.sendMessage(msgObj)
        if not result["OK"]:
            self.log.error("Cannot send Optimize message", result["Message"])
            return
        self.log.info("Optimize msg sent", "for %s jobs" % len(jids))

    ###########################################################################
    types_getMaxParametricJobs = []

    def export_getMaxParametricJobs(self):
        """Get the maximum number of parametric jobs

        :return: S_OK()/S_ERROR()
        """
        return S_OK(self.maxParametricJobs)

    types_submitJob = [six.string_types]

    def export_submitJob(self, jobDesc):
        """Submit a job to DIRAC WMS.
        The job can be a single job, or a parametric job.
        If it is a parametric job, then the parameters will need to be unpacked.

        :param str jobDesc: job description JDL (of a single or parametric job)
        :return: S_OK/S_ERROR, a list of newly created job IDs in case of S_OK.
        """

        if self.peerUsesLimitedProxy:
            return S_ERROR(EWMSSUBM, "Can't submit using a limited proxy")

        # Check job submission permission
        result = self.jobPolicy.getJobPolicy()
        if not result["OK"]:
            return S_ERROR(EWMSSUBM, "Failed to get job policies")
        policyDict = result["Value"]
        if not policyDict[RIGHT_SUBMIT]:
            return S_ERROR(EWMSSUBM, "Job submission not authorized")

        # jobDesc is JDL for now
        jobDesc = jobDesc.strip()
        if jobDesc[0] != "[":
            jobDesc = "[%s" % jobDesc
        if jobDesc[-1] != "]":
            jobDesc = "%s]" % jobDesc

        # Check if the job is a parametric one
        jobClassAd = ClassAd(jobDesc)
        result = getParameterVectorLength(jobClassAd)
        if not result["OK"]:
            self.log.error("Issue with getParameterVectorLength", result["Message"])
            return result
        nJobs = result["Value"]
        parametricJob = False
        if nJobs is not None and nJobs > 0:
            # if we are here, then jobDesc was the description of a parametric job. So we start unpacking
            parametricJob = True
            if nJobs > self.maxParametricJobs:
                self.log.error(
                    "Maximum of parametric jobs exceeded:",
                    "limit %d smaller than number of jobs %d" % (self.maxParametricJobs, nJobs),
                )
                return S_ERROR(EWMSJDL, "Number of parametric jobs exceeds the limit of %d" % self.maxParametricJobs)
            result = generateParametricJobs(jobClassAd)
            if not result["OK"]:
                return result
            jobDescList = result["Value"]
        else:
            # if we are here, then jobDesc was the description of a single job.
            jobDescList = [jobDesc]

        jobIDList = []

        if parametricJob:
            initialStatus = JobStatus.SUBMITTING
            initialMinorStatus = "Bulk transaction confirmation"
        else:
            initialStatus = JobStatus.RECEIVED
            initialMinorStatus = "Job accepted"

        for jobDescription in jobDescList:  # jobDescList because there might be a list generated by a parametric job
            result = self.jobDB.insertNewJobIntoDB(
                jobDescription,
                self.owner,
                self.ownerDN,
                self.ownerGroup,
                self.diracSetup,
                initialStatus=initialStatus,
                initialMinorStatus=initialMinorStatus,
            )
            if not result["OK"]:
                return result

            jobID = result["JobID"]
            self.log.info('Job added to the JobDB", "%s for %s/%s' % (jobID, self.ownerDN, self.ownerGroup))

            self.jobLoggingDB.addLoggingRecord(jobID, result["Status"], result["MinorStatus"], source="JobManager")

            jobIDList.append(jobID)

        # Set persistency flag
        retVal = gProxyManager.getUserPersistence(self.ownerDN, self.ownerGroup)
        if "Value" not in retVal or not retVal["Value"]:
            gProxyManager.setPersistency(self.ownerDN, self.ownerGroup, True)

        if parametricJob:
            result = S_OK(jobIDList)
        else:
            result = S_OK(jobIDList[0])

        result["JobID"] = result["Value"]
        result["requireProxyUpload"] = self.__checkIfProxyUploadIsRequired()
        # Ensure non-parametric jobs (i.e. non-bulk) get sent to optimizer immediately
        if not parametricJob:
            self.__sendJobsToOptimizationMind(jobIDList)
        return result

    ###########################################################################
    types_confirmBulkSubmission = [list]

    def export_confirmBulkSubmission(self, jobIDs):
        """Confirm the possibility to proceed with processing of the jobs specified
        by the jobIDList

        :param list jobIDs: list of job IDs

        :return: S_OK(list)/S_ERROR() -- confirmed job IDs
        """
        jobList = self.__getJobList(jobIDs)
        if not jobList:
            self.log.error("Issue with __getJobList", ": invalid job specification %s" % str(jobIDs))
            return S_ERROR(EWMSSUBM, "Invalid job specification: " + str(jobIDs))

        validJobList, _invalidJobList, _nonauthJobList, _ownerJobList = self.jobPolicy.evaluateJobRights(
            jobList, RIGHT_SUBMIT
        )

        # Check that all the requested jobs are eligible
        if set(jobList) != set(validJobList):
            return S_ERROR(EWMSSUBM, "Requested jobs for bulk transaction are not valid")

        result = self.jobDB.getJobsAttributes(jobList, ["Status", "MinorStatus"])
        if not result["OK"]:
            return S_ERROR(EWMSSUBM, "Requested jobs for bulk transaction are not valid")
        js_dict = strToIntDict(result["Value"])

        # Check if the jobs are already activated
        jobEnabledList = [
            jobID
            for jobID in jobList
            if js_dict[jobID]["Status"]
            in [JobStatus.RECEIVED, JobStatus.CHECKING, JobStatus.WAITING, JobStatus.MATCHED, JobStatus.RUNNING]
        ]
        if set(jobEnabledList) == set(jobList):
            return S_OK(jobList)

        # Check that requested job are in Submitting status
        jobUpdateStatusList = list(jobID for jobID in jobList if js_dict[jobID]["Status"] == JobStatus.SUBMITTING)
        if set(jobUpdateStatusList) != set(jobList):
            return S_ERROR(EWMSSUBM, "Requested jobs for bulk transaction are not valid")

        # Update status of all the requested jobs in one transaction
        result = self.jobDB.setJobAttributes(
            jobUpdateStatusList, ["Status", "MinorStatus"], [JobStatus.RECEIVED, "Job accepted"]
        )

        if not result["OK"]:
            return result

        self.__sendJobsToOptimizationMind(jobUpdateStatusList)
        return S_OK(jobUpdateStatusList)

    ###########################################################################
    def __checkIfProxyUploadIsRequired(self):
        """Check if an upload is required

        :return: bool
        """
        result = gProxyManager.userHasProxy(self.ownerDN, self.ownerGroup, validSeconds=18000)
        if not result["OK"]:
            self.log.error("Can't check if the user has proxy uploaded", result["Message"])
            return True
        # Check if an upload is required
        return not result["Value"]

    ###########################################################################

    @staticmethod
    def __getJobList(jobInput):
        """Evaluate the jobInput into a list of ints

        :param jobInput: one or more job IDs in int or str form
        :type jobInput: str or int or list
        :return : a list of int job IDs
        """

        if isinstance(jobInput, int):
            return [jobInput]
        if isinstance(jobInput, six.string_types):
            try:
                ijob = int(jobInput)
                return [ijob]
            except ValueError:
                return []
        if isinstance(jobInput, list):
            try:
                ljob = [int(x) for x in jobInput]
                return ljob
            except ValueError:
                return []

        return []

    ###########################################################################
    types_rescheduleJob = []

    def export_rescheduleJob(self, jobIDs):
        """Reschedule a single job. If the optional proxy parameter is given
        it will be used to refresh the proxy in the Proxy Repository

        :param list jobIDs: list of job IDs

        :return: S_OK()/S_ERROR() -- confirmed job IDs
        """

        jobList = self.__getJobList(jobIDs)
        if not jobList:
            return S_ERROR("Invalid job specification: " + str(jobIDs))

        validJobList, invalidJobList, nonauthJobList, ownerJobList = self.jobPolicy.evaluateJobRights(
            jobList, RIGHT_RESCHEDULE
        )
        for jobID in validJobList:
            self.taskQueueDB.deleteJob(jobID)
            # gJobDB.deleteJobFromQueue(jobID)
            result = self.jobDB.rescheduleJob(jobID)
            self.log.debug(str(result))
            if not result["OK"]:
                return result
            self.jobLoggingDB.addLoggingRecord(
                result["JobID"],
                status=result["Status"],
                minorStatus=result["MinorStatus"],
                applicationStatus="Unknown",
                source="JobManager",
            )

        if invalidJobList or nonauthJobList:
            result = S_ERROR("Some jobs failed reschedule")
            if invalidJobList:
                result["InvalidJobIDs"] = invalidJobList
            if nonauthJobList:
                result["NonauthorizedJobIDs"] = nonauthJobList
            return result

        result = S_OK(validJobList)
        result["requireProxyUpload"] = len(ownerJobList) > 0 and self.__checkIfProxyUploadIsRequired()
        self.__sendJobsToOptimizationMind(validJobList)
        return result

    types_removeJob = []

    def export_removeJob(self, jobIDs):
        """
        Completely remove a list of jobs, also from TaskQueueDB,
        and including its JobLogging info.
        Only authorized users are allowed to remove jobs.

        :param list jobIDs: list of job IDs
        :return: S_OK()/S_ERROR() -- confirmed job IDs
        """

        jobList = self.__getJobList(jobIDs)
        if not jobList:
            return S_ERROR("Invalid job specification: " + str(jobIDs))

        validJobList, invalidJobList, nonauthJobList, _ = self.jobPolicy.evaluateJobRights(jobList, RIGHT_DELETE)
        count = 0
        error_count = 0

        if validJobList:
            self.log.verbose("Removing jobs", "(n=%d)" % len(validJobList))
            result = self.jobDB.removeJobFromDB(validJobList)
            if not result["OK"]:
                self.log.error("Failed to remove jobs from JobDB", "(n=%d)" % len(validJobList))
            else:
                self.log.info("Removed jobs from JobDB", "(n=%d)" % len(validJobList))

            for jobID in validJobList:
                resultTQ = self.taskQueueDB.deleteJob(jobID)
                if not resultTQ["OK"]:
                    self.log.warn("Failed to remove job from TaskQueueDB", "(%d): %s" % (jobID, resultTQ["Message"]))
                    error_count += 1
                else:
                    count += 1

            result = self.jobLoggingDB.deleteJob(validJobList)
            if not result["OK"]:
                self.log.error("Failed to remove jobs from JobLoggingDB", "(n=%d)" % len(validJobList))
            else:
                self.log.info("Removed jobs from JobLoggingDB", "(n=%d)" % len(validJobList))

            if count > 0 or error_count > 0:
                self.log.info("Removed jobs from DB", "(%d jobs with %d errors)" % (count, error_count))

        if invalidJobList or nonauthJobList:
            self.log.error(
                "Jobs can not be removed",
                ": %d invalid and %d in nonauthJobList" % (len(invalidJobList), len(nonauthJobList)),
            )
            errMsg = "Some jobs failed removal"
            res = S_ERROR()
            if invalidJobList:
                self.log.debug("Invalid jobs: %s" % ",".join(str(ij) for ij in invalidJobList))
                res["InvalidJobIDs"] = invalidJobList
                errMsg += ": invalid jobs"
            if nonauthJobList:
                self.log.debug("nonauthJobList jobs: %s" % ",".join(str(nj) for nj in nonauthJobList))
                res["NonauthorizedJobIDs"] = nonauthJobList
                errMsg += ": non-authorized jobs"
            res["Message"] = errMsg
            return res

        return S_OK(validJobList)

    def __deleteJob(self, jobID):
        """Set the job status to "Deleted"
        and remove the pilot that ran and its logging info if the pilot is finished.

        :param int jobID: job ID
        :return: S_OK()/S_ERROR()
        """
        result = self.jobDB.setJobStatus(jobID, JobStatus.DELETED, "Checking accounting")
        if not result["OK"]:
            return result

        result = self.taskQueueDB.deleteJob(jobID)
        if not result["OK"]:
            self.log.warn("Failed to delete job from the TaskQueue")

        # if it was the last job for the pilot, clear PilotsLogging about it
        result = self.pilotAgentsDB.getPilotsForJobID(jobID)
        if not result["OK"]:
            self.log.error("Failed to get Pilots for JobID", result["Message"])
            return result
        for pilot in result["Value"]:
            res = self.pilotAgentsDB.getJobsForPilot(pilot)
            if not res["OK"]:
                self.log.error("Failed to get jobs for pilot", res["Message"])
                return res
            if not res["Value"]:  # if list of jobs for pilot is empty, delete pilot and pilotslogging
                result = self.pilotAgentsDB.getPilotInfo(pilotID=pilot)
                if not result["OK"]:
                    self.log.error("Failed to get pilot info", result["Message"])
                    return result
                pilotRef = result[0]["PilotJobReference"]
                ret = self.pilotAgentsDB.deletePilot(pilot)
                if not ret["OK"]:
                    self.log.error("Failed to delete pilot from PilotAgentsDB", ret["Message"])
                    return ret
                if self.pilotsLoggingDB:
                    ret = self.pilotsLoggingDB.deletePilotsLogging(pilotRef)
                    if not ret["OK"]:
                        self.log.error("Failed to delete pilot logging from PilotAgentsDB", ret["Message"])
                        return ret

        return S_OK()

    def __killJob(self, jobID, sendKillCommand=True):
        """Kill one job

        :param int jobID: job ID
        :param bool sendKillCommand: send kill command

        :return: S_OK()/S_ERROR()
        """
        if sendKillCommand:
            result = self.jobDB.setJobCommand(jobID, "Kill")
            if not result["OK"]:
                return result

        self.log.info("Job marked for termination", jobID)
        result = self.jobDB.setJobStatus(jobID, JobStatus.KILLED, "Marked for termination")
        if not result["OK"]:
            self.log.warn("Failed to set job Killed status", result["Message"])
        result = self.taskQueueDB.deleteJob(jobID)
        if not result["OK"]:
            self.log.warn("Failed to delete job from the TaskQueue", result["Message"])

        return S_OK()

    def __kill_delete_jobs(self, jobIDList, right):
        """Kill (== set the status to "KILLED") or delete (== set the status to "DELETED") jobs as necessary

        :param list jobIDList: job IDs
        :param str right: right

        :return: S_OK()/S_ERROR()
        """
        jobList = self.__getJobList(jobIDList)
        if not jobList:
            return S_ERROR("Invalid job specification: " + str(jobIDList))

        validJobList, invalidJobList, nonauthJobList, ownerJobList = self.jobPolicy.evaluateJobRights(jobList, right)

        badIDs = []

        if validJobList:
            # Get job status to see what is to be killed or deleted
            result = self.jobDB.getJobsAttributes(validJobList, ["Status"])
            if not result["OK"]:
                return result
            killJobList = []
            deleteJobList = []
            markKilledJobList = []
            stagingJobList = []
            for jobID, sDict in result["Value"].items():  # can be an iterator
                if sDict["Status"] in (JobStatus.RUNNING, JobStatus.MATCHED, JobStatus.STALLED):
                    killJobList.append(jobID)
                elif sDict["Status"] in (
                    JobStatus.SUBMITTING,
                    JobStatus.RECEIVED,
                    JobStatus.CHECKING,
                    JobStatus.WAITING,
                    JobStatus.RESCHEDULED,
                    JobStatus.DONE,
                    JobStatus.FAILED,
                    JobStatus.KILLED,
                ):
                    if not right == RIGHT_KILL:
                        deleteJobList.append(jobID)
                else:
                    markKilledJobList.append(jobID)
                if sDict["Status"] in [JobStatus.STAGING]:
                    stagingJobList.append(jobID)

            for jobID in markKilledJobList:
                result = self.__killJob(jobID, sendKillCommand=False)
                if not result["OK"]:
                    badIDs.append(jobID)

            for jobID in killJobList:
                result = self.__killJob(jobID)
                if not result["OK"]:
                    badIDs.append(jobID)

            for jobID in deleteJobList:
                result = self.__deleteJob(jobID)
                if not result["OK"]:
                    badIDs.append(jobID)

            if stagingJobList:
                stagerClient = StorageManagerClient()
                self.log.info("Going to send killing signal to stager as well!")
                result = stagerClient.killTasksBySourceTaskID(stagingJobList)
                if not result["OK"]:
                    self.log.warn("Failed to kill some Stager tasks", result["Message"])

        if nonauthJobList or badIDs:
            result = S_ERROR("Some jobs failed deletion")
            if nonauthJobList:
                self.log.warn("Non-authorized JobIDs won't be deleted", str(nonauthJobList))
                result["NonauthorizedJobIDs"] = nonauthJobList
            if badIDs:
                self.log.warn("JobIDs failed to be deleted", str(badIDs))
                result["FailedJobIDs"] = badIDs
            return result

        result = S_OK(validJobList)
        result["requireProxyUpload"] = len(ownerJobList) > 0 and self.__checkIfProxyUploadIsRequired()

        if invalidJobList:
            result["InvalidJobIDs"] = invalidJobList

        return result

    ###########################################################################
    types_deleteJob = []

    def export_deleteJob(self, jobIDs):
        """Delete jobs specified in the jobIDs list

        :param list jobIDs: list of job IDs

        :return: S_OK/S_ERROR
        """

        return self.__kill_delete_jobs(jobIDs, RIGHT_DELETE)

    ###########################################################################
    types_killJob = []

    def export_killJob(self, jobIDs):
        """Kill jobs specified in the jobIDs list

        :param list jobIDs: list of job IDs

        :return: S_OK/S_ERROR
        """

        return self.__kill_delete_jobs(jobIDs, RIGHT_KILL)

    ###########################################################################
    types_resetJob = []

    def export_resetJob(self, jobIDs):
        """Reset jobs specified in the jobIDs list

        :param list jobIDs: list of job IDs

        :return: S_OK/S_ERROR
        """

        jobList = self.__getJobList(jobIDs)
        if not jobList:
            return S_ERROR("Invalid job specification: " + str(jobIDs))

        validJobList, invalidJobList, nonauthJobList, ownerJobList = self.jobPolicy.evaluateJobRights(
            jobList, RIGHT_RESET
        )

        badIDs = []
        good_ids = []
        for jobID in validJobList:
            result = self.jobDB.setJobAttribute(jobID, "RescheduleCounter", -1)
            if not result["OK"]:
                badIDs.append(jobID)
            else:
                self.taskQueueDB.deleteJob(jobID)
                # gJobDB.deleteJobFromQueue(jobID)
                result = self.jobDB.rescheduleJob(jobID)
                if not result["OK"]:
                    badIDs.append(jobID)
                else:
                    good_ids.append(jobID)
                self.jobLoggingDB.addLoggingRecord(
                    result["JobID"],
                    status=result["Status"],
                    minorStatus=result["MinorStatus"],
                    applicationStatus="Unknown",
                    source="JobManager",
                )

        self.__sendJobsToOptimizationMind(good_ids)
        if invalidJobList or nonauthJobList or badIDs:
            result = S_ERROR("Some jobs failed resetting")
            if invalidJobList:
                result["InvalidJobIDs"] = invalidJobList
            if nonauthJobList:
                result["NonauthorizedJobIDs"] = nonauthJobList
            if badIDs:
                result["FailedJobIDs"] = badIDs
            return result

        result = S_OK()
        result["requireProxyUpload"] = len(ownerJobList) > 0 and self.__checkIfProxyUploadIsRequired()
        return result


class JobManagerHandler(JobManagerHandlerMixin, RequestHandler):
    def initialize(self):
        self.diracSetup = self.serviceInfoDict["clientSetup"]
        return self.initializeRequest()



#!/bin/env python
"""
tests for PoolComputingElement module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import time

import pytest

# sut
from DIRAC.Resources.Computing.PoolComputingElement import PoolComputingElement

jobScript = """#!/usr/bin/env python

from __future__ import print_function

import time
import os

jobNumber = %s
stopFile = 'stop_job_' + str(jobNumber)
start = time.time()

print("Start job", jobNumber, start)
while True:
  time.sleep(0.1)
  if os.path.isfile(stopFile):
    os.remove(stopFile)
    break
  if (time.time() - start) > 30:
    break
print("End job", jobNumber, time.time())
"""


def _stopJob(nJob):
    with open("stop_job_%s" % nJob, "w") as stopFile:
        stopFile.write("Stop")
    time.sleep(0.3)
    if os.path.isfile("stop_job_%s" % nJob):
        os.remove("stop_job_%s" % nJob)


@pytest.fixture
def createAndDelete():
    for i in range(6):
        with open("testPoolCEJob_%s.py" % i, "w") as execFile:
            execFile.write(jobScript % i)
        os.chmod("testPoolCEJob_%s.py" % i, 0o755)

    yield createAndDelete

    # from here on is teardown

    time.sleep(0.5)

    # stopping the jobs
    for i in range(6):
        _stopJob(i)

    # removing testPoolCEJob files
    # this will also stop the futures unless they are already stopped!
    for i in range(6):
        try:
            os.remove("testPoolCEJob_%s.py" % i)
        except OSError:
            pass


@pytest.mark.slow
def test_submit_and_shutdown(createAndDelete):

    time.sleep(0.5)

    ceParameters = {"WholeNode": True, "NumberOfProcessors": 4}
    ce = PoolComputingElement("TestPoolCE")
    ce.setParameters(ceParameters)

    result = ce.submitJob("testPoolCEJob_0.py", None)
    assert result["OK"] is True

    result = ce.shutdown()
    assert result["OK"] is True
    assert isinstance(result["Value"], dict)
    assert list(result["Value"].values())[0]["OK"] is True


def test_executeJob_wholeNode4(createAndDelete):

    time.sleep(0.5)

    ceParameters = {"WholeNode": True, "NumberOfProcessors": 4}
    ce = PoolComputingElement("TestPoolCE")
    ce.setParameters(ceParameters)

    # Test that max 4 processors can be used at a time
    result = ce.submitJob("testPoolCEJob_0.py", None)
    assert result["OK"] is True
    result = ce.getCEStatus()
    assert result["UsedProcessors"] == 1
    assert result["AvailableProcessors"] == 3
    assert result["RunningJobs"] == 1

    jobParams = {"mpTag": True, "numberOfProcessors": 2}
    result = ce.submitJob("testPoolCEJob_1.py", None, **jobParams)
    assert result["OK"] is True
    result = ce.getCEStatus()
    assert result["UsedProcessors"] == 3
    assert result["AvailableProcessors"] == 1
    assert result["RunningJobs"] == 2

    # now trying again would fail
    jobParams = {"mpTag": True, "numberOfProcessors": 2}
    result = ce.submitJob("testPoolCEJob_1.py", None, **jobParams)
    assert result["OK"] is False


@pytest.mark.slow
def test_executeJob_wholeNode8(createAndDelete):

    time.sleep(0.5)

    ceParameters = {"WholeNode": True, "NumberOfProcessors": 8}
    ce = PoolComputingElement("TestPoolCE")
    ce.setParameters(ceParameters)

    jobParams = {"mpTag": True, "numberOfProcessors": 2, "maxNumberOfProcessors": 2}
    result = ce.submitJob("testPoolCEJob_2.py", None, **jobParams)
    assert result["OK"] is True
    result = ce.getCEStatus()
    assert result["UsedProcessors"] == 2

    jobParams = {"mpTag": True, "numberOfProcessors": 1, "maxNumberOfProcessors": 3}
    result = ce.submitJob("testPoolCEJob_3.py", None, **jobParams)
    assert result["OK"] is True
    result = ce.getCEStatus()
    assert result["UsedProcessors"] == 5

    jobParams = {"numberOfProcessors": 2}  # This is same as asking for SP
    result = ce.submitJob("testPoolCEJob_4.py", None, **jobParams)
    assert result["OK"] is True
    result = ce.getCEStatus()
    assert result["UsedProcessors"] == 6

    # now trying again would fail
    jobParams = {"mpTag": True, "numberOfProcessors": 3}
    result = ce.submitJob("testPoolCEJob_5.py", None, **jobParams)
    assert result["OK"] is False
    assert "Not enough processors" in result["Message"]

    # waiting 40 seconds and then submit again
    time.sleep(40)
    jobParams = {"mpTag": True, "numberOfProcessors": 3}
    result = ce.submitJob("testPoolCEJob_5.py", None, **jobParams)
    assert result["OK"] is True
    time.sleep(10)

    result = ce.shutdown()
    assert result["OK"] is True
    assert isinstance(result["Value"], dict)
    assert len(result["Value"]) == 4


def test_executeJob_submitAndStop(createAndDelete):

    time.sleep(0.5)

    ceParameters = {"WholeNode": True, "NumberOfProcessors": 4}
    ce = PoolComputingElement("TestPoolCE")
    ce.setParameters(ceParameters)

    jobParams = {"mpTag": True, "numberOfProcessors": 2, "maxNumberOfProcessors": 2}
    result = ce.submitJob("testPoolCEJob_0.py", None, **jobParams)
    assert result["OK"] is True
    result = ce.getCEStatus()
    assert result["UsedProcessors"] == 2
    assert result["AvailableProcessors"] == 2
    assert result["RunningJobs"] == 1

    time.sleep(5)
    _stopJob(0)
    # Allow job to stop
    time.sleep(2)

    result = ce.getCEStatus()
    assert result["RunningJobs"] == 0
    assert result["UsedProcessors"] == 0
    assert result["AvailableProcessors"] == 4


@pytest.mark.slow
def test_executeJob_WholeNodeJobs(createAndDelete):

    time.sleep(0.5)

    ce = PoolComputingElement("TestPoolCE")
    ceParameters = {"WholeNode": False, "NumberOfProcessors": 4}
    ce.setParameters(ceParameters)

    jobParams = {"mpTag": True, "numberOfProcessors": 2, "maxNumberOfProcessors": 2}
    result = ce.submitJob("testPoolCEJob_0.py", None, **jobParams)
    assert result["OK"] is True

    jobParams = {"mpTag": True, "numberOfProcessors": 2}
    result = ce.submitJob("testPoolCEJob_5.py", None, **jobParams)
    assert result["OK"] is True
    result = ce.getCEStatus()
    assert result["UsedProcessors"] == 4
    assert result["AvailableProcessors"] == 0
    assert result["RunningJobs"] == 2

    # Allow jobs to start, then stopping them
    time.sleep(5)
    for i in range(8):
        _stopJob(i)
    # Allow jobs to stop
    time.sleep(2)

    result = ce.getCEStatus()
    assert result["UsedProcessors"] == 0

    # Trying with whole node jobs
    result = ce.submitJob("testPoolCEJob_0.py", None)  # first 1 SP job
    assert result["OK"] is True
    result = ce.getCEStatus()
    assert result["UsedProcessors"] == 1

    jobParams = {"mpTag": True, "wholeNode": True}
    result = ce.submitJob("testPoolCEJob_1.py", None, **jobParams)
    assert result["OK"] is False
    assert "Not enough processors for the job" in result["Message"]
    # Allow job to start
    time.sleep(5)

    _stopJob(0)
    # Allow job to stop
    time.sleep(2)

    jobParams = {"mpTag": True, "wholeNode": True}
    result = ce.submitJob("testPoolCEJob_1.py", None, **jobParams)
    assert result["OK"] is True
    result = ce.getCEStatus()
    assert result["UsedProcessors"] == 4


@pytest.mark.parametrize(
    "processorsPerTask, kwargs, expected",
    [
        (None, {}, 1),
        (None, {"mpTag": False}, 1),
        (None, {"mpTag": True}, 1),
        (None, {"mpTag": True, "wholeNode": True}, 16),
        (None, {"mpTag": True, "wholeNode": False}, 1),
        (None, {"mpTag": True, "numberOfProcessors": 4}, 4),
        (None, {"mpTag": True, "numberOfProcessors": 4, "maxNumberOfProcessors": 8}, 8),
        (None, {"mpTag": True, "numberOfProcessors": 4, "maxNumberOfProcessors": 32}, 16),
        ({1: 4}, {"mpTag": True, "wholeNode": True}, 0),
        ({1: 4}, {"mpTag": True, "wholeNode": False}, 1),
        ({1: 4}, {"mpTag": True, "numberOfProcessors": 2}, 2),
        ({1: 4}, {"mpTag": True, "maxNumberOfProcessors": 2}, 2),
        ({1: 4}, {"mpTag": True, "maxNumberOfProcessors": 16}, 12),
    ],
)
def test__getProcessorsForJobs(processorsPerTask, kwargs, expected):
    ce = PoolComputingElement("TestPoolCE")
    ce.processors = 16

    if processorsPerTask:
        ce.processorsPerTask = processorsPerTask
    res = ce._getProcessorsForJobs(kwargs)
    assert res == expected



#!/usr/bin/env python

"""
Get the files attached to a transformation
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument("transID: transformation ID")
    _, args = Script.parseCommandLine()

    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    if len(args) != 1:
        Script.showHelp(exitCode=1)

    tc = TransformationClient()
    res = tc.getTransformationFiles({"TransformationID": args[0]})

    if not res["OK"]:
        DIRAC.gLogger.error(res["Message"])
        DIRAC.exit(2)

    for transfile in res["Value"]:
        DIRAC.gLogger.notice(transfile["LFN"])


if __name__ == "__main__":
    main()



#!/usr/bin/env python
########################################################################
# File :   dirac-platform
# Author : Adria Casajus
########################################################################
"""
The *dirac-platform* script determines the "platform" of a certain node.
The platform is a string used to identify the minimal characteristics of the node,
enough to determine which version of DIRAC can be installed.

Invoked at any installation, so by the *dirac-install* script, and by the pilots.

On a RHEL 6 node, for example, the determined dirac platform is "Linux_x86_64_glibc-2.5"

Usage:
  dirac-platform [options]

Example:
  $ dirac-platform
  Linux_x86_64_glibc-2.5

"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"


try:
    from DIRAC.Core.Utilities.Platform import getPlatformString
except Exception:
    import argparse
    import platform
    import os
    import sys
    import re
    import subprocess

    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.parse_known_args()

    # We need to patch python platform module. It does a string comparison for the libc versions.
    # it fails when going from 2.9 to 2.10,
    # the fix converts the version to a tuple and attempts a numeric comparison

    _libc_search = re.compile(r"(__libc_init)" "|" "(GLIBC_([0-9.]+))" "|" r"(libc(_\w+)?\.so(?:\.(\d[0-9.]*))?)")

    def libc_ver(executable=sys.executable, lib="", version="", chunksize=2048):
        """Tries to determine the libc version that the file executable
        (which defaults to the Python interpreter) is linked against.

        Returns a tuple of strings (lib,version) which default to the
        given parameters in case the lookup fails.

        Note that the function has intimate knowledge of how different
        libc versions add symbols to the executable and thus is probably
        only useable for executables compiled using gcc.

        The file is read and scanned in chunks of chunksize bytes.

        """
        with open(executable, "rb") as f:
            binary = f.read(chunksize)
            pos = 0
            version = [0, 0, 0]
            while True:
                m = _libc_search.search(binary, pos)
                if not m:
                    binary = f.read(chunksize)
                    if not binary:
                        break
                    pos = 0
                    continue
                libcinit, glibc, glibcversion, so, threads, soversion = m.groups()
                if libcinit and not lib:
                    lib = "libc"
                elif glibc:
                    glibcversion_parts = glibcversion.split(".")
                    for i in range(len(glibcversion_parts)):
                        try:
                            glibcversion_parts[i] = int(glibcversion_parts[i])
                        except ValueError:
                            glibcversion_parts[i] = 0
                    if libcinit and not lib:
                        lib = "libc"
                    elif glibc:
                        if lib != "glibc":
                            lib = "glibc"
                            version = glibcversion_parts
                        elif glibcversion_parts > version:
                            version = glibcversion_parts
                elif so:
                    if lib != "glibc":
                        lib = "libc"
                        version = max(version, soversion)
                        if threads and version[-len(threads) :] != threads:
                            version = version + threads
                pos = m.end()
        return lib, ".".join(map(str, version))

    # Command line interface
    def getPlatformString():
        # Modified to return our desired platform string, R. Graciani
        platformTuple = (platform.system(), platform.machine())
        if platformTuple[0] == "Linux":
            sp = subprocess.Popen(["/sbin/ldconfig", "--print-cache"], stdout=subprocess.PIPE)
            ldre = re.compile(r".*=> (.*/libc\.so\..*$)")
            libs = []
            for line in sp.stdout.readlines():
                reM = ldre.match(line)
                if reM:
                    libs.append(reM.groups()[0])
            if not libs:
                # get version of higher libc installed
                if platform.machine().find("64") != -1:
                    lib = "/lib64"
                else:
                    lib = "/lib"
                for libFile in os.listdir(lib):
                    if libFile.find("libc-") == 0 or libFile.find("libc.so") == 0:
                        libs.append(os.path.join(lib, libFile))
            newest_lib = [0, 0, 0]
            for lib in libs:
                lib_parts = libc_ver(lib)[1].split(".")
                for i in range(len(lib_parts)):
                    try:
                        lib_parts[i] = int(lib_parts[i])
                    except ValueError:
                        lib_parts[i] = 0
                        # print "non integer version numbers"
                if lib_parts > newest_lib:
                    newest_lib = lib_parts

            platformTuple += ("glibc-" + ".".join(map(str, newest_lib)),)
        elif platformTuple[0] == "Darwin":
            platformTuple += (".".join(platform.mac_ver()[0].split(".")[:2]),)
        elif platformTuple[0] == "Windows":
            platformTuple += (platform.win32_ver()[0],)
        else:
            platformTuple += platform.release()

        platformString = "%s_%s_%s" % platformTuple

        return platformString

    def main():
        print(getPlatformString())


else:
    from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

    @Script()
    def main():
        print(getPlatformString())


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Get the given file replica metadata from the File Catalog
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os

from DIRAC import exit as DIRACExit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(("LocalFile: Path to local file containing LFNs", "LFN:       Logical File Names"))
    Script.registerArgument(" SE:        Storage element")
    Script.parseCommandLine()

    from DIRAC import gLogger
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    inputFileName, storageElement = Script.getPositionalArgs(group=True)

    if os.path.exists(inputFileName):
        inputFile = open(inputFileName, "r")
        string = inputFile.read()
        lfns = [lfn.strip() for lfn in string.splitlines()]
        inputFile.close()
    else:
        lfns = [inputFileName]

    res = DataManager().getReplicaMetadata(lfns, storageElement)
    if not res["OK"]:
        print("Error:", res["Message"])
        DIRACExit(1)

    print("%s %s %s %s" % ("File".ljust(100), "Migrated".ljust(8), "Cached".ljust(8), "Size (bytes)".ljust(10)))
    for lfn, metadata in res["Value"]["Successful"].items():
        print(
            "%s %s %s %s"
            % (
                lfn.ljust(100),
                str(metadata["Migrated"]).ljust(8),
                str(metadata.get("Cached", metadata["Accessible"])).ljust(8),
                str(metadata["Size"]).ljust(10),
            )
        )
    for lfn, reason in res["Value"]["Failed"].items():
        print("%s %s" % (lfn.ljust(100), reason.ljust(8)))


if __name__ == "__main__":
    main()



"""TransformationInfo class to be used by ILCTransformation System"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from collections import OrderedDict, defaultdict
from six.moves import zip_longest

from DIRAC import gLogger, S_OK
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.Proxy import UserProxy
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.TransformationSystem.Utilities.JobInfo import JobInfo
from DIRAC.WorkloadManagementSystem.Client import JobStatus
from DIRAC.WorkloadManagementSystem.Client.JobStateUpdateClient import JobStateUpdateClient

__RCSID__ = "$Id$"


class TransformationInfo(object):
    """Hold information about a transformation."""

    def __init__(self, transformationID, transInfoDict, enabled, tClient, fcClient, jobMon):
        """Store clients etc."""
        self.log = gLogger.getSubLogger(__name__ + "[%s]" % transformationID)
        self.enabled = enabled
        self.tID = transformationID
        self.transName = transInfoDict["TransformationName"]
        self.tClient = tClient
        self.jobMon = jobMon
        self.fcClient = fcClient
        self.transType = transInfoDict["Type"]
        self.authorDN = transInfoDict["AuthorDN"]
        self.authorGroup = transInfoDict["AuthorGroup"]
        self.jobStateClient = JobStateUpdateClient()

    def checkTasksStatus(self):
        """Check the status for the task of given transformation and taskID"""

        res = self.tClient.getTransformationFiles(condDict={"TransformationID": self.tID})
        if not res["OK"]:
            raise RuntimeError("Failed to get transformation tasks: %s" % res["Message"])

        tasksDict = defaultdict(list)
        for task in res["Value"]:
            taskID = task["TaskID"]
            lfn = task["LFN"]
            status = task["Status"]
            fileID = task["FileID"]
            errorCount = task["ErrorCount"]
            tasksDict[taskID].append(dict(FileID=fileID, LFN=lfn, Status=status, ErrorCount=errorCount))

        return tasksDict

    def setJobDone(self, job):
        """set the taskID to Done"""
        if not self.enabled:
            return
        self.__setTaskStatus(job, "Done")
        if job.status != JobStatus.DONE:
            self.__updateJobStatus(job.jobID, JobStatus.DONE, "Job forced to Done")

    def setJobFailed(self, job):
        """set the taskID to Done"""
        if not self.enabled:
            return
        self.__setTaskStatus(job, "Failed")
        if job.status != JobStatus.FAILED:
            self.__updateJobStatus(job.jobID, JobStatus.FAILED, "Job forced to Failed")

    def setInputUnused(self, job):
        """Set the inputfiles to unused"""
        self.__setInputStatus(job, "Unused")

    def setInputMaxReset(self, job):
        """set the inputfile to MaxReset"""
        self.__setInputStatus(job, "MaxReset")

    def setInputProcessed(self, job):
        """set the inputfile to processed"""
        self.__setInputStatus(job, "Processed")

    def setInputDeleted(self, job):
        """set the inputfile to processed"""
        self.__setInputStatus(job, "Deleted")

    def __setInputStatus(self, job, status):
        """set the input file to status"""
        if self.enabled:
            result = self.tClient.setFileStatusForTransformation(self.tID, status, job.inputFiles, force=True)
            if not result["OK"]:
                gLogger.error("Failed updating status", result["Message"])
                raise RuntimeError("Failed updating file status")

    def __setTaskStatus(self, job, status):
        """update the task in the TransformationDB"""
        taskID = job.taskID
        res = self.tClient.setTaskStatus(self.transName, taskID, status)
        if not res["OK"]:
            raise RuntimeError("Failed updating task status: %s" % res["Message"])

    def __updateJobStatus(self, jobID, status, minorstatus=""):
        """Update the job status."""
        if self.enabled:
            source = "DataRecoveryAgent"
            result = self.jobStateClient.setJobStatus(jobID, status, minorstatus, source, None, True)
        else:
            return S_OK("DisabledMode")
        if not result["OK"]:
            self.log.error("Failed to update job status", result["Message"])
            raise RuntimeError("Failed to update job status")
        return result

    def __findAllDescendants(self, lfnList):
        """Find all descendants of a list of LFNs"""
        allDescendants = []
        result = self.fcClient.getFileDescendents(lfnList, list(range(1, 8)))
        if not result["OK"]:
            return allDescendants
        for dummy_lfn, descendants in result["Value"]["Successful"].items():
            allDescendants.extend(descendants)
        return allDescendants

    def cleanOutputs(self, jobInfo):
        """Remove all job outputs for job represented by jobInfo object.

        Including removal of descendents, if defined.
        """
        if len(jobInfo.outputFiles) == 0:
            return
        descendants = self.__findAllDescendants(jobInfo.outputFiles)
        existingOutputFiles = [
            lfn for lfn, status in zip_longest(jobInfo.outputFiles, jobInfo.outputFileStatus) if status == "Exists"
        ]
        filesToDelete = existingOutputFiles + descendants

        if not filesToDelete:
            return

        if not self.enabled:
            self.log.notice("Would have removed these files: \n +++ %s " % "\n +++ ".join(filesToDelete))
            return
        self.log.notice("Remove these files: \n +++ %s " % "\n +++ ".join(filesToDelete))

        errorReasons = defaultdict(list)
        successfullyRemoved = 0

        for lfnList in breakListIntoChunks(filesToDelete, 200):
            with UserProxy(proxyUserDN=self.authorDN, proxyUserGroup=self.authorGroup) as proxyResult:
                if not proxyResult["OK"]:
                    raise RuntimeError("Failed to get a proxy: %s" % proxyResult["Message"])
                result = DataManager().removeFile(lfnList)
                if not result["OK"]:
                    self.log.error("Failed to remove LFNs", result["Message"])
                    raise RuntimeError("Failed to remove LFNs: %s" % result["Message"])
                for lfn, err in result["Value"]["Failed"].items():
                    reason = str(err)
                    errorReasons[reason].append(lfn)
                successfullyRemoved += len(result["Value"]["Successful"])
        for reason, lfns in errorReasons.items():
            self.log.error("Failed to remove %d files with error: %s" % (len(lfns), reason))
        self.log.notice("Successfully removed %d files" % successfullyRemoved)

    def getJobs(self, statusList=None):
        """Get done and failed jobs.

        :param list statusList: optional list of status to find jobs
        :returns: 3-tuple of OrderedDict of JobInfo objects, keyed by jobID;
                  number of Done jobs; number of Failed jobs
        """
        done = S_OK([])
        failed = S_OK([])
        if statusList is None:
            statusList = [JobStatus.DONE, JobStatus.FAILED]
        if "Done" in statusList:
            self.log.notice("Getting 'Done' Jobs...")
            done = self.__getJobs([JobStatus.DONE])
        if "Failed" in statusList:
            self.log.notice("Getting 'Failed' Jobs...")
            failed = self.__getJobs([JobStatus.FAILED])
        done = done["Value"]
        failed = failed["Value"]

        jobsUnsorted = {}
        for job in done:
            jobsUnsorted[int(job)] = JobInfo(job, JobStatus.DONE, self.tID, self.transType)
        for job in failed:
            jobsUnsorted[int(job)] = JobInfo(job, JobStatus.FAILED, self.tID, self.transType)
        jobs = OrderedDict(sorted(jobsUnsorted.items(), key=lambda t: t[0]))

        self.log.notice("Found %d Done Jobs " % len(done))
        self.log.notice("Found %d Failed Jobs " % len(failed))
        return jobs, len(done), len(failed)

    def __getJobs(self, status):
        """Return list of jobs with given status.

        :param list status: list of status to find
        :returns: S_OK with result
        :raises: RuntimeError when failing to find jobs
        """
        attrDict = dict(Status=status, JobGroup="%08d" % int(self.tID))
        res = self.jobMon.getJobs(attrDict)
        if res["OK"]:
            self.log.debug("Found Trans jobs: %s" % res["Value"])
            return res
        else:
            self.log.error("Error finding jobs: ", res["Message"])
            raise RuntimeError("Failed to get jobs")



""" CEAvailabilityPolicy module
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.ResourceStatusSystem.PolicySystem.PolicyBase import PolicyBase


class CEAvailabilityPolicy(PolicyBase):
    """
    The CEAvailabilityPolicy checks if the CE is in 'Production' or not on the BDII.
    """

    @staticmethod
    def _evaluate(commandResult):
        """
        It returns Active status if CE is in 'Production'.
        Banned if the CE is different from 'Production'.

        commandResult is a dictionary like:
          {'OK': True,
          'Value': {
            'Reason': "All queues in 'Production'",
            'Status': 'Production',
            'cccreamceli05.in2p3.fr:8443/cream-sge-long': 'Production',
            'cccreamceli05.in2p3.fr:8443/cream-sge-verylong': 'Production'
            }
          }

        Otherwise, it returns error.
        """

        result = {"Status": None, "Reason": None}

        if not commandResult["OK"]:
            result["Status"] = "Error"
            result["Reason"] = commandResult["Message"]
            return S_OK(result)

        commandResult = commandResult["Value"]

        if commandResult["Status"] == "Production":
            result["Status"] = "Active"
        else:
            result["Status"] = "Banned"

        result["Reason"] = commandResult["Reason"]

        return S_OK(result)
